from unittest.mock import patch, mock_open

import pytest

from refinitiv.data.content.esg.bulk._actions import parse_log_file
from .conftest import raise_


def test_actions_get_created_tables(actions):
    # when
    tables = actions.get_created_tables()

    # then
    assert tables is not None


def test_actions_add_pass_dict(actions):
    # given
    input_details = {"details": "details"}

    # when
    details = actions.add("action_name", **input_details)

    # then
    assert details == input_details


def test_actions_add_pass_kwargs(actions):
    # given
    expected_details = {"details": "details"}

    # when
    details = actions.add("action_name", details="details")

    # then
    assert details == expected_details


def test_actions_cleanup(actions):
    try:
        # when
        actions.cleanup()
    except Exception as e:
        assert False, str(e)


@pytest.mark.parametrize(
    "exception",
    [
        PermissionError,
        FileNotFoundError,
    ],
)
def test_actions_cleanup_when_error(actions, exception):
    # given
    remove_file_mock = patch("os.remove", new=lambda *_: raise_(exception))
    file_mock = patch("builtins.open", new=mock_open())

    # when
    file_mock.start()
    remove_file_mock.start()
    result = actions.cleanup()
    remove_file_mock.stop()
    file_mock.stop()

    # then
    assert result is None


def test_actions_details_by_action(actions):
    # when
    details_by_action = actions.details_by_action

    # then
    assert details_by_action is not None


def test_actions_downloaded(actions):
    # when
    downloaded = actions.downloaded()

    # then
    assert downloaded is not None


def test_actions_extracted(actions):
    # when
    extracted = actions.extracted()

    # then
    assert extracted is not None


def test_actions_get_downloaded(actions):
    # when
    downloaded = actions.get_downloaded()

    # then
    assert downloaded is not None


def test_actions_get_extracted(actions):
    # when
    extracted = actions.get_extracted()

    # then
    assert extracted is not None


def test_actions_get_not_updated_and_extracted(actions):
    # when
    not_updated, extracted = actions.get_not_updated_and_extracted()

    # then
    assert not_updated is not None and extracted is not None


def test_actions_table_created(actions):
    # when
    created = actions.table_created()

    # then
    assert created is not None


def test_actions_updated(actions):
    # when
    updated = actions.updated()

    # then
    assert updated is not None


def test_actions_cleaned_up_db(actions):
    # when
    cleaned_up = actions.cleaned_up_db()

    # then
    assert cleaned_up is not None


def test_parse_log_file():
    # given
    logpath = "./tests/unit/content/esg/bulk/test_log.txt"

    # when
    details_by_log_action = parse_log_file(logpath)

    # then
    assert details_by_log_action


def test_parse_log_file_catch_exception():
    # given
    logpath = "invalid_test_log.txt"

    # when
    details_by_log_action = parse_log_file(logpath)

    # then
    assert details_by_log_action is not None


def test_actions_update(actions):
    # given
    actions._details_by_action = ["1", "2ewf"]

    # when
    actions.update()

    # then
    assert actions._details_by_action is None



# coding: utf8
from .._object_definition import ObjectDefinition

from ._enums import EtiInputVolatilityType
from ._enums import VolatilityModel
from ._enums import PriceSide
from ._enums import TimeStamp
from ._enums import MoneynessType
from ._enums import Axis
from ._models import MoneynessWeight
from ._models import SurfaceFilters


class EtiSurfaceParameters(ObjectDefinition):
    def __init__(
        self,
        filters=None,
        input_volatility_type=None,
        moneyness_type=None,
        price_side=None,
        time_stamp=None,
        volatility_model=None,
        weights=None,
        x_axis=None,
        y_axis=None,
        calculation_date=None,
        svi_alpha_extrapolation=None,
    ):
        super().__init__()
        self.filters = filters
        self.input_volatility_type = input_volatility_type
        self.moneyness_type = moneyness_type
        self.price_side = price_side
        self.time_stamp = time_stamp
        self.volatility_model = volatility_model
        self.weights = weights
        self.x_axis = x_axis
        self.y_axis = y_axis
        self.calculation_date = calculation_date
        self.svi_alpha_extrapolation = svi_alpha_extrapolation

    @property
    def filters(self):
        """
        The details of the filtering
        :return: object SurfaceFilters
        """
        return self._get_object_parameter(SurfaceFilters, "filters")

    @filters.setter
    def filters(self, value):
        self._set_object_parameter(SurfaceFilters, "filters", value)

    @property
    def input_volatility_type(self):
        """
        Specifies the type of volatility used as an input of the model (calculated Implied Volatility, Settlement)

         - Settle: [DEPRECATED] The service uses the settlement volatility to build the volatility surface

         - Quoted: The service uses the quoted volatility to build the volatility surface

         - Implied: The service internally calculates implied volatilities for the option universe before building the surface

        Default value is "Implied".
        :return: enum EtiInputVolatilityType
        """
        return self._get_enum_parameter(EtiInputVolatilityType, "inputVolatilityType")

    @input_volatility_type.setter
    def input_volatility_type(self, value):
        self._set_enum_parameter(EtiInputVolatilityType, "inputVolatilityType", value)

    @property
    def moneyness_type(self):
        """
        The enumerate that specifies the moneyness type to use for calibration.
        - Spot
        - Fwd
        - Sigma
        Optional. Default value is "Spot".
        :return: enum MoneynessType
        """
        return self._get_enum_parameter(MoneynessType, "moneynessType")

    @moneyness_type.setter
    def moneyness_type(self, value):
        self._set_enum_parameter(MoneynessType, "moneynessType", value)

    @property
    def price_side(self):
        """
        Specifies whether bid, ask or mid is used to build the surface.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def time_stamp(self):
        """
        Define how the timestamp is selected:
        - Open: the opening value of the valuationDate or if not available the close of the previous day is used.
        - Default: the latest snapshot is used when valuationDate is today, the close price when valuationDate is in the past.
        :return: enum TimeStamp
        """
        return self._get_enum_parameter(TimeStamp, "timeStamp")

    @time_stamp.setter
    def time_stamp(self, value):
        self._set_enum_parameter(TimeStamp, "timeStamp", value)

    @property
    def volatility_model(self):
        """
        The quantitative model used to generate the volatility surface. This may depend on the asset class.
        :return: enum VolatilityModel
        """
        return self._get_enum_parameter(VolatilityModel, "volatilityModel")

    @volatility_model.setter
    def volatility_model(self, value):
        self._set_enum_parameter(VolatilityModel, "volatilityModel", value)

    @property
    def weights(self):
        """
        Specifies the list of calibration weight.
        :return: list MoneynessWeight
        """
        return self._get_list_parameter(MoneynessWeight, "weights")

    @weights.setter
    def weights(self, value):
        self._set_list_parameter(MoneynessWeight, "weights", value)

    @property
    def x_axis(self):
        """
        Specifies the unit for the x axis (e.g. Date, Tenor)
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "xAxis")

    @x_axis.setter
    def x_axis(self, value):
        self._set_enum_parameter(Axis, "xAxis", value)

    @property
    def y_axis(self):
        """
        Specifies the unit for the y axis (e.g. Strike, Delta). This may depend on the asset class.
        For Fx Volatility Surface, we support both Delta and Strike.
        :return: enum Axis
        """
        return self._get_enum_parameter(Axis, "yAxis")

    @y_axis.setter
    def y_axis(self, value):
        self._set_enum_parameter(Axis, "yAxis", value)

    @property
    def calculation_date(self):
        """
        The date the volatility surface is generated.
        :return: str
        """
        return self._get_parameter("calculationDate")

    @calculation_date.setter
    def calculation_date(self, value):
        self._set_parameter("calculationDate", value)

    @property
    def svi_alpha_extrapolation(self):
        """
        Svi Alpha Extrapolation for building the surface
        Default value : TRUE
        :return: bool
        """
        return self._get_parameter("sviAlphaExtrapolation")

    @svi_alpha_extrapolation.setter
    def svi_alpha_extrapolation(self, value):
        self._set_parameter("sviAlphaExtrapolation", value)



from .._data._data_provider_layer import DataProviderLayer
from .._data._data_type import DataType


class CFSStream(DataProviderLayer):
    def __init__(self, id):
        super().__init__(data_type=DataType.CFS_STREAM, id=id)



# coding: utf8

from typing import Optional, TYPE_CHECKING

from ._bermudan_swaption_definition import BermudanSwaptionDefinition
from ._swaption_definition import SwaptionInstrumentDefinition
from ._swaption_pricing_parameters import PricingParameters
from .. import swap
from .._base_definition import BaseDefinition
from ..._enums import (
    BuySell,
    ExerciseStyle,
    SwaptionSettlementType,
    PremiumSettlementType,
    SwaptionType,
)
from ..._models import InputFlow
from ....._tools import create_repr, try_copy_to_list

if TYPE_CHECKING:
    from ....._types import ExtendedParams, OptStrStrs


class Definition(BaseDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        A user defined string to identify the instrument. it can be used to link output
        results to the instrument definition.limited to 40 characters.only alphabetic,
        numeric and '- _.#=@' characters are supported. optional. no default value
        applies.
    start_date : str, optional
        The date the swaption starts. optional. by default it is derived from the
        tradedate and the day to spot convention of the contract currency.
    end_date : str, optional
        The maturity or expiry date of the instrument's leg. the value is expressed in
        iso 8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g. 2021-01-01t00:00:00z). either
        tenor or enddate must be provided. the default value is valuationdate shifted
        forward by tenor.
    tenor : str, optional
        The code indicating the period between startdate and enddate of the instrument
        (e.g. '6m', '1y'). mandatory, if enddate is not provided. the default value is
        calculated from enddate.
    notional_amount : float, optional
        The notional amount of the instrument. the default value is '1,000,000'.
    bermudan_swaption_definition : BermudanSwaptionDefinition, optional

    buy_sell : BuySell, optional
        The indicator of the deal side. the possible values are:   buy: buying the
        option,   sell: selling/writing the option.  no default value applies.
    exercise_style : ExerciseStyle, optional
        The option style based on its exercise restrictions. the possible values are:
        amer,   euro,   berm.  note: all exercise styles may not apply to certain option
        no default value applies.
    payments : InputFlow, optional
        An array of payments
    premium_settlement_type : PremiumSettlementType, optional
        The cash settlement type of the option premium   spot,   forward.
    settlement_type : SwaptionSettlementType, optional
        The settlement method for options when exercised. the possible values are:
        physical: delivering the underlying asset, or for a swaption, physically
        entering into the underlying swap.    cash: paying out in cash.  the default
        value is 'physical'.
    swaption_type : SwaptionType, optional
        The indicator if the swaption is a payer or a receiver. the possible values are:
        receiver: a right to receive a fixed rate of the underlying swap,   payer: a
        right to pay a fixed rate of the underlying swap.  no default value applies.
    underlying_definition : SwapDefinition, optional

    spread_vs_atm_in_bp : float, optional
        Spread between strike and atm strike, expressed in basis points (bp).
    strike_percent : float, optional
        The set price at which the owner of the option can buy or sell the underlying
        asset. for a swaption, it is the fixed rate of the underlying swap at which the
        owner of the swaption can enter the swap. the value is expressed in percentages.
        by default, fixedratepercent of the underlying swap is used.
    fields: list of str, optional
        Contains the list of Analytics that the quantitative analytic service will
        compute.
    pricing_parameters : PricingParameters, optional
        The pricing parameters to apply to this instrument. If pricing parameters
        are not provided at this level parameters defined globally at the request
        level are used. If no pricing parameters are provided globally default
        values apply.
    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response, async_mode=None)
        Returns a response to the data platform
    get_data_async(session=session, on_response=on_response, async_mode=None)
        Returns a response to the async data platform
    get_stream(session=session)
        Get stream quantitative analytic service subscription

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> definition = rdf.swaption.Definition(
     ...   buy_sell=rdf.swaption.BuySell.BUY,
     ...   call_put=rdf.swaption.CallPut.CALL,
     ...   exercise_style=rdf.swaption.ExerciseStyle.BERM,
     ...   underlying_definition=rdf.swap.Definition(tenor="3Y", template="EUR_AB6E"),
     ...)
     >>> response = definition.get_data()

    Using get_stream
     >>> stream = definition.get_stream()
     >>> stream.open()
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_amount: Optional[float] = None,
        bermudan_swaption_definition: Optional[BermudanSwaptionDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        payments: Optional[InputFlow] = None,
        premium_settlement_type: Optional[PremiumSettlementType] = None,
        settlement_type: Optional[SwaptionSettlementType] = None,
        swaption_type: Optional[SwaptionType] = None,
        underlying_definition: Optional[swap.Definition] = None,
        spread_vs_atm_in_bp: Optional[float] = None,
        strike_percent: Optional[float] = None,
        fields: "OptStrStrs" = None,
        pricing_parameters: Optional[PricingParameters] = None,
        extended_params: "ExtendedParams" = None,
    ):
        fields = try_copy_to_list(fields)
        self.underlying_definition = underlying_definition
        definition = SwaptionInstrumentDefinition(
            bermudan_swaption_definition=bermudan_swaption_definition,
            buy_sell=buy_sell,
            exercise_style=exercise_style,
            payments=payments,
            premium_settlement_type=premium_settlement_type,
            settlement_type=settlement_type,
            swaption_type=swaption_type,
            underlying_definition=underlying_definition,
            end_date=end_date,
            instrument_tag=instrument_tag,
            notional_amount=notional_amount,
            spread_vs_atm_in_bp=spread_vs_atm_in_bp,
            start_date=start_date,
            strike_percent=strike_percent,
            tenor=tenor,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.ipa.financial_contracts.swaption",
            content=f"{{underlying_definition='{self.underlying_definition}'}}",
        )



import logging
from threading import Thread
from unittest.mock import MagicMock, patch
import refinitiv.data as rd
import pytest

from refinitiv.data import _configure
from refinitiv.data._core.session import (
    RDDefaultSessionManager,
    EikonDefaultSessionManager,
    _rd_default_session_manager,
    _eikon_default_session_manager,
    DesktopSession,
    Session,
    PlatformSession,
)


def test_rd_default_session_manager_try_get_default_session_error():
    # given
    manager = RDDefaultSessionManager()

    with pytest.raises(
        AttributeError,
        match="No default session created yet. Please create a session first!",
    ):
        manager.try_get_default_session()


def test_rd_default_session_manager_try_get_default_session():
    # given
    mock_session = MagicMock()
    manager = RDDefaultSessionManager(mock_session)

    result = manager.try_get_default_session()

    assert result == mock_session


def test_rd_default_session_manager_set_default_session_error():
    manager = RDDefaultSessionManager()

    with pytest.raises(TypeError, match="Invalid argument"):
        manager.set_default_session("invalid-type")


def test_rd_default_session_manager_close_default_session():
    mock_session = MagicMock()
    manager = RDDefaultSessionManager(mock_session)

    manager.close_default_session()

    mock_session.close.assert_called_once()


def test_rd_default_session_manager_clear_default_session():
    mock_session = MagicMock()
    manager = RDDefaultSessionManager(mock_session)

    manager.set_default_session(None)

    assert manager.has_session() is False


def test_rd_default_session_manager_set_default_session_set_session():
    manager = RDDefaultSessionManager()

    manager.set_default_session(DesktopSession(""))

    assert manager.has_session() is True


def test_eikon_default_session_manager_get_default_session():
    manager = EikonDefaultSessionManager()

    default_session = manager.get_default_session()

    assert default_session is None


def test_eikon_default_session_manager_get_default_session_set_session():
    manager = EikonDefaultSessionManager()

    default_session = manager.get_default_session("mocked-app-key")

    assert default_session is not None


def test_eikon_default_session_manager_set_default_session_set_session():
    manager = EikonDefaultSessionManager()

    manager.set_default_session(DesktopSession(""))

    assert manager.has_session() is True


@patch("refinitiv.data._core.session.DesktopSession")
def test_eikon_default_session_manager_get_session_with_app_key(mock_desktop_session):
    mock_session = MagicMock()
    manager = EikonDefaultSessionManager(mock_session)

    default_session = manager.get_default_session("mocked-app-key")

    assert default_session is not None


def test_eikon_default_session_manager_close_default_session():
    mock_session = MagicMock()
    manager = EikonDefaultSessionManager(mock_session)

    manager.close_default_session()

    mock_session.close.assert_called_once()


def test_eikon_default_session_manager_clear_default_session():
    mock_session = MagicMock()
    manager = EikonDefaultSessionManager(mock_session)

    manager.clear_default_session()

    assert manager.has_session() is False


def test_if_rd_manager_clear_session_in_eikon_manager_session_is_empty():
    _rd_default_session_manager.set_default_session(DesktopSession(""))

    _rd_default_session_manager.set_default_session(None)

    assert _eikon_default_session_manager.has_session() is False


def test_if_eikon_manager_clear_session_in_rd_manager_session_is_empty():
    _eikon_default_session_manager.set_default_session(DesktopSession(""))

    _eikon_default_session_manager.clear_default_session()

    assert _rd_default_session_manager.has_session() is False


def test_get_log_level_return_default_config_value_if_not_set(session):
    # given
    expected_level = _configure.defaults.log_level

    # when
    testing_level = session.get_log_level()

    # then
    assert testing_level == expected_level


def test_get_log_level_return_int_value(session):
    # when
    testing_level = session.get_log_level()

    # then
    assert isinstance(testing_level, int)


def test_get_log_level_returns_same_that_pass_in_set_log_level(session):
    # given
    expected_level = 1

    # when
    session.set_log_level(expected_level)
    testing_level = session.get_log_level()

    # then
    assert testing_level == expected_level


def test_set_log_level_return_none(session):
    # given
    input_level = 1

    # when
    testing_value = session.set_log_level(input_level)

    # then
    assert testing_value is None


@pytest.mark.parametrize(
    "input_level,expected_level",
    [
        [logging.DEBUG, logging.DEBUG],
        [logging.INFO, logging.INFO],
        [logging.WARNING, logging.WARNING],
        [logging.ERROR, logging.ERROR],
        [logging.FATAL, logging.FATAL],
        [logging.CRITICAL, logging.CRITICAL],
    ],
)
def test_set_logging_levels_as_int_and_get_same_int_levels(
    input_level, expected_level, session
):
    # given
    session.set_log_level(input_level)

    # when
    testing_level = session.get_log_level()

    # then
    assert testing_level == expected_level and isinstance(testing_level, int)


@pytest.mark.parametrize(
    "input_level,expected_level",
    [
        ("DEBUG", 10),
        ("INFO", 20),
        ("WARNING", 30),
        ("ERROR", 40),
        ("FATAL", 50),
        ("CRITICAL", 50),
    ],
)
def test_set_logging_levels_as_string_and_get_same_int_levels(
    input_level, expected_level, session
):
    # given
    session.set_log_level(input_level)

    # when
    testing_level = session.get_log_level()

    # then
    assert testing_level == expected_level and isinstance(testing_level, int)


@pytest.mark.parametrize(
    "session",
    [
        PlatformSession(
            app_key="app_key", deployed_platform_host="deployed_platform_host"
        ),
        DesktopSession(app_key="app_key"),
    ],
    ids=["PlatformSession", "DesktopSession"],
)
def test_session_has_pending_state(session):
    def check_state():
        nonlocal testing_state
        testing_state = session.open_state

    def open():
        while not stop:
            pass

    # given
    testing_state = None
    expected_state = rd.OpenState.Pending
    stop = False
    session._connection.open = open

    # when
    Thread(target=session.open, daemon=True).start()
    Thread(target=check_state, daemon=True).start()

    # then
    stop = True
    assert testing_state == expected_state, testing_state



from typing import Optional, TYPE_CHECKING

from ......_tools import try_copy_to_list
from .....ipa._enums._extrapolation_mode import ExtrapolationMode
from ...._curves._enums import CompoundingType
from ...._object_definition import ObjectDefinition
from ..._bond_curves._enums import (
    BasisSplineSmoothModel,
    CalendarAdjustment,
    CalibrationModel,
    InterestCalculationMethod,
    InterpolationMode,
    PriceSide,
)


if TYPE_CHECKING:
    from ......_types import OptStrings, OptInt, OptBool, OptStr


class CreditCurveParameters(ObjectDefinition):
    """
    Generates the credit curves for the definition provided

    Parameters
    ----------
    interest_calculation_method : InterestCalculationMethod, optional
        The day count basis method used to compute the points of the zero coupon curve.
        the possible values are:   * dcb_30_360     actual number of days in the coupon
        period calculated on the basis of a year of 360 days with twelve 30-day months
        unless:     - the last day of the period is the 31st day of a month and the
        first day of the period is a day other than the 30th or 31st day of a month, in
        which case the month that includes the last day shall not be considered to be
        shortened to a 30-day month,     - the last day of the period is the last day of
        the month of february, in which case the month of february shall not be
        considered to be lengthened to a 30-day month.   * dcb_30_actual     the day
        count is identical to 30/360 (us) and the year basis is identical to
        actual/actual.   * dcb_actual_360     the day count is the actual number of days
        of the period. the year basis is 360.   * dcb_actual_365     the day count is
        the actual number of days of the period. the year basis is 365.   *
        dcb_actual_actual     the dcb is calculated by actual days / year basis where:
        - actual days are defined as the actual days between the starting date
        (d1.m1.y1) and end date (d2.m2.y2).     - year basis is defined as the actual
        days between the start date (d1.m1.y1) and the next relevant interest payment
        date (d3.m3.y3) multiplied by the instrument coupon frequency.   *
        dcb_actual_actual_isda     similar to actual/365, except for a period that
        includes days falling in a leap year. it is calculated by dcb = number of days
        in a leap year/366 + number of days in a non-leap year/365.     a convention is
        also known as actual/365 isda.   *dcb_30_360_isda     for two dates (y1,m1,d1)
        and (y2,m2,d2):     - if d1 is 31, change it to 30,     - if d2 is 31 and d1 is
        30, change d2 to 30.     then the date difference is
        (y2-y1)x360+(m2-m1)*30+(d2-d1).   * dcb_30_365_isda     for two dates (y1,m1,d1)
        and (y2,m2,d2):     - if d1=31 then d1=30,     - if d2=31 and d1=30 or 31 then
        d2=30.     then the date difference is (y2-y1)*365+(m2-m1)*30+(d2-d1)   *
        dcb_30_360_us     for two dates (y1,m1,d1) and (y2,m2,d2):     - if d1=31 then
        d1=30,     - if d2=31 and d1=30 or 31 then d2=30,     - if d1 is the last day of
        february then d1=30,     - if d1 is the last day of february and d2 is the last
        day of february then d2=30.     the last day of february is february 29 in leap
        years and february 28 in non leap years.     the 30/360 us rule is identical to
        30/360 isda when the eom (end-of-month) convention does not apply. this
        indicates whether all coupon payment dates fall on the last day of the month. if
        the investment is not eom, it will always pay on the same day of the month
        (e.g., the 10th).     then the date difference is
        (y2-y1)x360+(m2-m1)*30+(d2-d1).   * dcb_actual_actual_afb     the dcb is
        calculated by actual days / year basis where:     actual days are defined as the
        actual days between the start date (d1.m1.y1) and end date (d2.m2.y2).     year
        basis is either 365 if the calculation period does not contain 29th feb, or 366
        if the calculation period includes 29th feb.   * dcb_workingdays_252     the day
        count is the actual number of business days of the period according to the
        instrument calendars. the year basis is 252. commonly used in the brazilian
        market.   * dcb_actual_365l     the day count is the actual number of days of
        the period. the year basis is calculated in the following two rules:     - if
        the coupon frequency is annual, then year basis is 366 if the 29 feb. is
        included in the interest period, else 365,     - if the coupon frequency is not
        annual, then year basis is 366 for each interest period where ending date falls
        in a leap year, otherwise it is 365.   * dcb_actualleapday_365     the day count
        ignores 29th february when counting days. the year basis is 365 days.   *
        dcb_actualleapday_360     the day count ignores 29th february when counting
        days. the year basis is 360 days.   * dcb_actual_36525     the day count is the
        actual number of days of the period. the year basis is 365.25.   *
        dcb_actual_365_canadianconvention     follows the canadian domestic bond market
        convention. the day count basis is computed as follows:     - if the number of
        days of a period is less than the actual number of days in a regular coupon
        period the dcb_actual_365 convention is used,     - otherwise: dcb = 1 -
        daysremaininginperiod x frequency / 365.   * dcb_30_360_german     for two dates
        (y1,m1,d1) and (y2,m2,d2):     - if d1=31 then d1=30,     - if d2=31 then d2=30,
        - if d1 is the last day of february then d1=30,     - if d2 is the last day of
        february then d2=30.     the last day of february is february 29 in leap years
        and february 28 in non leap years.     then the date difference is
        (y2-y1)x360+(m2-m1)*30+(d2-d1).   * dcb_30_365_german     similar to 30/360
        (german), except that the year basis is treated as 365 days.   *
        dcb_30_actual_german     the day count is identical to 30/360 (german) and the
        year basis is similar to actual/actual. this method was formerly used in the
        eurobond markets.   * dcb_30e_360_isma     actual number of days in the coupon
        period calculated on the basis of a year of 360 days with twelve 30-day months
        (regardless of the date of the first day or last day of the period).   *
        dcb_actual_364     a special case of actual/actual (isma) when a coupon period
        contains 91 or 182 days. actual/364 applies for some short-term instruments.
        day count basis = 364.   * dcb_30_actual_isda   * dcb_30_365_brazil   *
        dcb_actual_365p   * dcb_constant
    basis_spline_smooth_model : BasisSplineSmoothModel, optional
        Basis spline model. values can be: - mccullochlinearregression -
        waggonersmoothingsplinemodel - andersonsmoothingsplinemodel
    calendar_adjustment : CalendarAdjustment, optional
        The cash flow adjustment according to a selected calendar. the possible values
        are:   * no   * weekend: for the cash flow pricing using the calendar 'weekend'
        * calendar: for the cash flow pricing using the calendar defined by the
        parameter 'calendars'. the default value is 'calendar'.
    calendars : string, optional
        The list of comma-separated calendar codes used to define non-working days and
        to adjust interest rate curve coupon dates and values (e.g., 'emu_fi'). by
        default, the calendar code is derived from the interest rate curve currency.
    calibration_model : CalibrationModel, optional
        Bond zero coupon curve calibration method. values can be:   - basisspline   -
        nelsonsiegelsvensson   - bootstrap
    compounding_type : CompoundingType, optional
        The yield type of the interest rate curve. the possible values are:   *
        continuous   * moneymarket   * compounded   * discounted the default value is
        'compounded'.
    extrapolation_mode : ExtrapolationMode, optional
        The extrapolation method used in the zero coupon curve bootstrapping. the
        possible values are:   * none: no extrapolation,   * constant: constant
        extrapolation,   * linear: linear extrapolation. the default value is 'none'.
    interpolation_mode : InterpolationMode, optional
        The interpolation method used in zero curve bootstrapping. the possible values
        are:   * cubicdiscount: local cubic interpolation of discount factors   *
        cubicrate: local cubic interpolation of rates   * cubicspline: a natural cubic
        spline   * forwardmonotoneconvex: forward monotone convexc interpolation   *
        linear: linear interpolation   * log: log-linear interpolation   * hermite:
        hermite (bessel) interpolation   * akimamethod: the akima method (a smoother
        variant of local cubic interpolation)   * fritschbutlandmethod: the
        fritsch-butland method (a monotonic cubic variant)   * krugermethod: the kruger
        method (a monotonic cubic variant)   * monotoniccubicnaturalspline: a monotonic
        natural cubic spline   * monotonichermitecubic: monotonic hermite (bessel) cubic
        interpolation   * tensionspline: a tension spline
    price_side : PriceSide, optional
        The quoted price side of the instrument to be used for the zero coupon curve
        construction. the possible values are:   * bid   * ask   * mid the default value
        is 'mid'.
    basis_spline_knots : int, optional
        Number of knots you can choose to build the yield curve when using the
        basis-spline models
    return_calibrated_parameters : bool, optional
        If true, returns parametric model calibrated parameters
    use_delayed_data_if_denied : bool, optional
        Use delayed ric to retrieve data when not permissioned on constituent ric. the
        default value is 'false'.
    use_duration_weighted_minimization : bool, optional
        Specifies the type of minimization of residual errors in the vasicek-fong model
        and basis spline model: - true: minimize residual errors between market and
        model prices weighted by the inverse of the modified duration of the bonds -
        false: minimize residual errors between market and model prices
    use_multi_dimensional_solver : bool, optional
        An indicator whether the multi-dimensional solver for yield curve bootstrapping
        must be used. this solving method is required because the bootstrapping method
        sometimes creates a zero coupon curve which does not accurately reprice the
        input instruments used to build it. the multi-dimensional solver is recommended
        when cubic interpolation methods are used in building the curve (in other cases,
        performance might be inferior to the regular bootstrapping method). when use for
        credit curve it is only used when the calibrationmodel is set to 'bootstrap'.
        the possible values are:   * true: the multi-dimensional solver is used,   *
        false: the multi-dimensional solver is not used. the default value is 'true'.
    valuation_date : str, optional
        Valuation date for this curve, that means the data at which instrument market
        data is retrieved.
    """

    def __init__(
        self,
        interest_calculation_method: Optional[InterestCalculationMethod] = None,
        basis_spline_smooth_model: Optional[BasisSplineSmoothModel] = None,
        calendar_adjustment: Optional[CalendarAdjustment] = None,
        calendars: "OptStrings" = None,
        calibration_model: Optional[CalibrationModel] = None,
        compounding_type: Optional[CompoundingType] = None,
        extrapolation_mode: Optional[ExtrapolationMode] = None,
        interpolation_mode: Optional[InterpolationMode] = None,
        price_side: Optional[PriceSide] = None,
        basis_spline_knots: "OptInt" = None,
        return_calibrated_parameters: "OptBool" = None,
        use_delayed_data_if_denied: "OptBool" = None,
        use_duration_weighted_minimization: "OptBool" = None,
        use_multi_dimensional_solver: "OptBool" = None,
        valuation_date: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.interest_calculation_method = interest_calculation_method
        self.basis_spline_smooth_model = basis_spline_smooth_model
        self.calendar_adjustment = calendar_adjustment
        self.calendars = try_copy_to_list(calendars)
        self.calibration_model = calibration_model
        self.compounding_type = compounding_type
        self.extrapolation_mode = extrapolation_mode
        self.interpolation_mode = interpolation_mode
        self.price_side = price_side
        self.basis_spline_knots = basis_spline_knots
        self.return_calibrated_parameters = return_calibrated_parameters
        self.use_delayed_data_if_denied = use_delayed_data_if_denied
        self.use_duration_weighted_minimization = use_duration_weighted_minimization
        self.use_multi_dimensional_solver = use_multi_dimensional_solver
        self.valuation_date = valuation_date

    @property
    def basis_spline_smooth_model(self):
        """
        Basis spline model. values can be: - mccullochlinearregression -
        waggonersmoothingsplinemodel - andersonsmoothingsplinemodel
        :return: enum BasisSplineSmoothModel
        """
        return self._get_enum_parameter(
            BasisSplineSmoothModel, "basisSplineSmoothModel"
        )

    @basis_spline_smooth_model.setter
    def basis_spline_smooth_model(self, value):
        self._set_enum_parameter(
            BasisSplineSmoothModel, "basisSplineSmoothModel", value
        )

    @property
    def calendar_adjustment(self):
        """
        The cash flow adjustment according to a selected calendar. the possible values
        are:   * no   * weekend: for the cash flow pricing using the calendar 'weekend'
        * calendar: for the cash flow pricing using the calendar defined by the
        parameter 'calendars'. the default value is 'calendar'.
        :return: enum CalendarAdjustment
        """
        return self._get_enum_parameter(CalendarAdjustment, "calendarAdjustment")

    @calendar_adjustment.setter
    def calendar_adjustment(self, value):
        self._set_enum_parameter(CalendarAdjustment, "calendarAdjustment", value)

    @property
    def calendars(self):
        """
        The list of comma-separated calendar codes used to define non-working days and
        to adjust interest rate curve coupon dates and values (e.g., 'emu_fi'). by
        default, the calendar code is derived from the interest rate curve currency.
        :return: list string
        """
        return self._get_list_parameter(str, "calendars")

    @calendars.setter
    def calendars(self, value):
        self._set_list_parameter(str, "calendars", value)

    @property
    def calibration_model(self):
        """
        Bond zero coupon curve calibration method. values can be:   - basisspline   -
        nelsonsiegelsvensson   - bootstrap
        :return: enum CalibrationModel
        """
        return self._get_enum_parameter(CalibrationModel, "calibrationModel")

    @calibration_model.setter
    def calibration_model(self, value):
        self._set_enum_parameter(CalibrationModel, "calibrationModel", value)

    @property
    def compounding_type(self):
        """
        The yield type of the interest rate curve. the possible values are:   *
        continuous   * moneymarket   * compounded   * discounted the default value is
        'compounded'.
        :return: enum CompoundingType
        """
        return self._get_enum_parameter(CompoundingType, "compoundingType")

    @compounding_type.setter
    def compounding_type(self, value):
        self._set_enum_parameter(CompoundingType, "compoundingType", value)

    @property
    def extrapolation_mode(self):
        """
        The extrapolation method used in the zero coupon curve bootstrapping. the
        possible values are:   * none: no extrapolation,   * constant: constant
        extrapolation,   * linear: linear extrapolation. the default value is 'none'.
        :return: enum ExtrapolationMode
        """
        return self._get_enum_parameter(ExtrapolationMode, "extrapolationMode")

    @extrapolation_mode.setter
    def extrapolation_mode(self, value):
        self._set_enum_parameter(ExtrapolationMode, "extrapolationMode", value)

    @property
    def interest_calculation_method(self):
        """
        The day count basis method used to compute the points of the zero coupon curve.
        the possible values are:   * dcb_30_360     actual number of days in the coupon
        period calculated on the basis of a year of 360 days with twelve 30-day months
        unless:     - the last day of the period is the 31st day of a month and the
        first day of the period is a day other than the 30th or 31st day of a month, in
        which case the month that includes the last day shall not be considered to be
        shortened to a 30-day month,     - the last day of the period is the last day of
        the month of february, in which case the month of february shall not be
        considered to be lengthened to a 30-day month.   * dcb_30_actual     the day
        count is identical to 30/360 (us) and the year basis is identical to
        actual/actual.   * dcb_actual_360     the day count is the actual number of days
        of the period. the year basis is 360.   * dcb_actual_365     the day count is
        the actual number of days of the period. the year basis is 365.   *
        dcb_actual_actual     the dcb is calculated by actual days / year basis where:
        - actual days are defined as the actual days between the starting date
        (d1.m1.y1) and end date (d2.m2.y2).     - year basis is defined as the actual
        days between the start date (d1.m1.y1) and the next relevant interest payment
        date (d3.m3.y3) multiplied by the instrument coupon frequency.   *
        dcb_actual_actual_isda     similar to actual/365, except for a period that
        includes days falling in a leap year. it is calculated by dcb = number of days
        in a leap year/366 + number of days in a non-leap year/365.     a convention is
        also known as actual/365 isda.   *dcb_30_360_isda     for two dates (y1,m1,d1)
        and (y2,m2,d2):     - if d1 is 31, change it to 30,     - if d2 is 31 and d1 is
        30, change d2 to 30.     then the date difference is
        (y2-y1)x360+(m2-m1)*30+(d2-d1).   * dcb_30_365_isda     for two dates (y1,m1,d1)
        and (y2,m2,d2):     - if d1=31 then d1=30,     - if d2=31 and d1=30 or 31 then
        d2=30.     then the date difference is (y2-y1)*365+(m2-m1)*30+(d2-d1)   *
        dcb_30_360_us     for two dates (y1,m1,d1) and (y2,m2,d2):     - if d1=31 then
        d1=30,     - if d2=31 and d1=30 or 31 then d2=30,     - if d1 is the last day of
        february then d1=30,     - if d1 is the last day of february and d2 is the last
        day of february then d2=30.     the last day of february is february 29 in leap
        years and february 28 in non leap years.     the 30/360 us rule is identical to
        30/360 isda when the eom (end-of-month) convention does not apply. this
        indicates whether all coupon payment dates fall on the last day of the month. if
        the investment is not eom, it will always pay on the same day of the month
        (e.g., the 10th).     then the date difference is
        (y2-y1)x360+(m2-m1)*30+(d2-d1).   * dcb_actual_actual_afb     the dcb is
        calculated by actual days / year basis where:     actual days are defined as the
        actual days between the start date (d1.m1.y1) and end date (d2.m2.y2).     year
        basis is either 365 if the calculation period does not contain 29th feb, or 366
        if the calculation period includes 29th feb.   * dcb_workingdays_252     the day
        count is the actual number of business days of the period according to the
        instrument calendars. the year basis is 252. commonly used in the brazilian
        market.   * dcb_actual_365l     the day count is the actual number of days of
        the period. the year basis is calculated in the following two rules:     - if
        the coupon frequency is annual, then year basis is 366 if the 29 feb. is
        included in the interest period, else 365,     - if the coupon frequency is not
        annual, then year basis is 366 for each interest period where ending date falls
        in a leap year, otherwise it is 365.   * dcb_actualleapday_365     the day count
        ignores 29th february when counting days. the year basis is 365 days.   *
        dcb_actualleapday_360     the day count ignores 29th february when counting
        days. the year basis is 360 days.   * dcb_actual_36525     the day count is the
        actual number of days of the period. the year basis is 365.25.   *
        dcb_actual_365_canadianconvention     follows the canadian domestic bond market
        convention. the day count basis is computed as follows:     - if the number of
        days of a period is less than the actual number of days in a regular coupon
        period the dcb_actual_365 convention is used,     - otherwise: dcb = 1 -
        daysremaininginperiod x frequency / 365.   * dcb_30_360_german     for two dates
        (y1,m1,d1) and (y2,m2,d2):     - if d1=31 then d1=30,     - if d2=31 then d2=30,
        - if d1 is the last day of february then d1=30,     - if d2 is the last day of
        february then d2=30.     the last day of february is february 29 in leap years
        and february 28 in non leap years.     then the date difference is
        (y2-y1)x360+(m2-m1)*30+(d2-d1).   * dcb_30_365_german     similar to 30/360
        (german), except that the year basis is treated as 365 days.   *
        dcb_30_actual_german     the day count is identical to 30/360 (german) and the
        year basis is similar to actual/actual. this method was formerly used in the
        eurobond markets.   * dcb_30e_360_isma     actual number of days in the coupon
        period calculated on the basis of a year of 360 days with twelve 30-day months
        (regardless of the date of the first day or last day of the period).   *
        dcb_actual_364     a special case of actual/actual (isma) when a coupon period
        contains 91 or 182 days. actual/364 applies for some short-term instruments.
        day count basis = 364.   * dcb_30_actual_isda   * dcb_30_365_brazil   *
        dcb_actual_365p   * dcb_constant
        :return: enum InterestCalculationMethod
        """
        return self._get_enum_parameter(
            InterestCalculationMethod, "interestCalculationMethod"
        )

    @interest_calculation_method.setter
    def interest_calculation_method(self, value):
        self._set_enum_parameter(
            InterestCalculationMethod, "interestCalculationMethod", value
        )

    @property
    def interpolation_mode(self):
        """
        The interpolation method used in zero curve bootstrapping. the possible values
        are:   * cubicdiscount: local cubic interpolation of discount factors   *
        cubicrate: local cubic interpolation of rates   * cubicspline: a natural cubic
        spline   * forwardmonotoneconvex: forward monotone convexc interpolation   *
        linear: linear interpolation   * log: log-linear interpolation   * hermite:
        hermite (bessel) interpolation   * akimamethod: the akima method (a smoother
        variant of local cubic interpolation)   * fritschbutlandmethod: the
        fritsch-butland method (a monotonic cubic variant)   * krugermethod: the kruger
        method (a monotonic cubic variant)   * monotoniccubicnaturalspline: a monotonic
        natural cubic spline   * monotonichermitecubic: monotonic hermite (bessel) cubic
        interpolation   * tensionspline: a tension spline
        :return: enum InterpolationMode
        """
        return self._get_enum_parameter(InterpolationMode, "interpolationMode")

    @interpolation_mode.setter
    def interpolation_mode(self, value):
        self._set_enum_parameter(InterpolationMode, "interpolationMode", value)

    @property
    def price_side(self):
        """
        The quoted price side of the instrument to be used for the zero coupon curve
        construction. the possible values are:   * bid   * ask   * mid the default value
        is 'mid'.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def basis_spline_knots(self):
        """
        Number of knots you can choose to build the yield curve when using the
        basis-spline models
        :return: int
        """
        return self._get_parameter("basisSplineKnots")

    @basis_spline_knots.setter
    def basis_spline_knots(self, value):
        self._set_parameter("basisSplineKnots", value)

    @property
    def return_calibrated_parameters(self):
        """
        If true, returns parametric model calibrated parameters
        :return: bool
        """
        return self._get_parameter("returnCalibratedParameters")

    @return_calibrated_parameters.setter
    def return_calibrated_parameters(self, value):
        self._set_parameter("returnCalibratedParameters", value)

    @property
    def use_delayed_data_if_denied(self):
        """
        Use delayed ric to retrieve data when not permissioned on constituent ric. the
        default value is 'false'.
        :return: bool
        """
        return self._get_parameter("useDelayedDataIfDenied")

    @use_delayed_data_if_denied.setter
    def use_delayed_data_if_denied(self, value):
        self._set_parameter("useDelayedDataIfDenied", value)

    @property
    def use_duration_weighted_minimization(self):
        """
        Specifies the type of minimization of residual errors in the vasicek-fong model
        and basis spline model: - true: minimize residual errors between market and
        model prices weighted by the inverse of the modified duration of the bonds -
        false: minimize residual errors between market and model prices
        :return: bool
        """
        return self._get_parameter("useDurationWeightedMinimization")

    @use_duration_weighted_minimization.setter
    def use_duration_weighted_minimization(self, value):
        self._set_parameter("useDurationWeightedMinimization", value)

    @property
    def use_multi_dimensional_solver(self):
        """
        An indicator whether the multi-dimensional solver for yield curve bootstrapping
        must be used. this solving method is required because the bootstrapping method
        sometimes creates a zero coupon curve which does not accurately reprice the
        input instruments used to build it. the multi-dimensional solver is recommended
        when cubic interpolation methods are used in building the curve (in other cases,
        performance might be inferior to the regular bootstrapping method). when use for
        credit curve it is only used when the calibrationmodel is set to 'bootstrap'.
        the possible values are:   * true: the multi-dimensional solver is used,   *
        false: the multi-dimensional solver is not used. the default value is 'true'.
        :return: bool
        """
        return self._get_parameter("useMultiDimensionalSolver")

    @use_multi_dimensional_solver.setter
    def use_multi_dimensional_solver(self, value):
        self._set_parameter("useMultiDimensionalSolver", value)

    @property
    def valuation_date(self):
        """
        Valuation date for this curve, that means the data at which instrument market
        data is retrieved.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



"""
        python-configuration module.
        https://github.com/tr11/python-configuration
        version = 0.8.3
"""

import json
import os
from importlib.abc import InspectLoader
from types import ModuleType
from typing import Any, Dict, Iterable, List, Optional, TextIO, Union, cast

try:
    import yaml
except ImportError:  # pragma: no cover
    yaml = None
try:
    import toml
except ImportError:  # pragma: no cover
    toml = None


from .configuration import Configuration
from .configuration_set import ConfigurationSet
from .helpers import InterpolateEnumType, InterpolateType


def config(
    *configs: Iterable,
    prefix: str = "",
    separator: Optional[str] = None,
    remove_level: int = 1,
    lowercase_keys: bool = False,
    ignore_missing_paths: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> ConfigurationSet:
    """
    Create a :class:`ConfigurationSet` instance from an iterable of configs.

    :param configs: iterable of configurations
    :param prefix: prefix to filter environment variables with
    :param remove_level: how many levels to remove from the resulting config
    :param lowercase_keys: whether to convert every key to lower case.
    :param ignore_missing_paths: whether to ignore failures from missing files/folders.
    :param separator: separator for Python modules and environment variables.
    :param interpolate: whether to apply string interpolation when looking for items

    Note that the :py:attr:`separator` parameter  impacts Python modules and and
    environment variables at the same time. To pass different separators to Python
    modules and environments, use the longer version
    ``('python', 'path-to-module', prefix, separator)``
    and ``('env', prefix, separator)`` .
    """
    instances = []
    default_args: List[str] = [prefix]
    if separator is not None:
        default_args.append(separator)
    default_kwargs: Dict[Any, Any] = {
        "lowercase_keys": lowercase_keys,
        # for Configuration Sets, interpolate parameters should be at the Set level
        "interpolate": False,
        "interpolate_type": InterpolateEnumType.STANDARD,
    }

    for config_ in configs:
        if isinstance(config_, dict):
            instances.append(config_from_dict(config_, **default_kwargs))
            continue
        elif isinstance(config_, str):
            if config_.endswith(".py"):
                config_ = ("python", config_, *default_args)
            elif config_.endswith(".json"):
                config_ = ("json", config_, True)
            elif yaml and config_.endswith(".yaml"):
                config_ = ("yaml", config_, True)
            elif toml and config_.endswith(".toml"):
                config_ = ("toml", config_, True)
            elif config_.endswith(".ini"):
                config_ = ("ini", config_, True)
            elif config_.endswith(".env"):
                config_ = ("dotenv", config_, True)
            elif os.path.isdir(config_):
                config_ = ("path", config_, remove_level)
            elif config_ in ("env", "environment"):
                config_ = ("env", *default_args)
            elif all(s and s.isidentifier() for s in config_.split(".")):
                config_ = ("python", config_, *default_args)
            else:
                raise ValueError(f'Cannot determine config type from "{config_}"')

        if not isinstance(config_, (tuple, list)) or len(config_) == 0:
            raise ValueError(
                "configuration parameters must be a list of dictionaries,"
                " strings, or non-empty tuples/lists"
            )
        type_ = config_[0]
        if type_ == "dict":
            instances.append(config_from_dict(*config_[1:], **default_kwargs))
        elif type_ in ("env", "environment"):
            params = list(config_[1:]) + default_args[(len(config_) - 1) :]
            instances.append(config_from_env(*params, **default_kwargs))
        elif type_ == "python":
            if len(config_) < 2:
                raise ValueError("No path specified for python module")
            params = list(config_[1:]) + default_args[(len(config_) - 2) :]
            try:
                instances.append(config_from_python(*params, **default_kwargs))
            except (FileNotFoundError, ModuleNotFoundError):
                if not ignore_missing_paths:
                    raise
        elif type_ == "json":
            try:
                instances.append(config_from_json(*config_[1:], **default_kwargs))
            except FileNotFoundError:
                if not ignore_missing_paths:
                    raise
        elif yaml and type_ == "yaml":
            try:
                instances.append(config_from_yaml(*config_[1:], **default_kwargs))
            except FileNotFoundError:
                if not ignore_missing_paths:
                    raise
        elif toml and type_ == "toml":
            try:
                instances.append(config_from_toml(*config_[1:], **default_kwargs))
            except FileNotFoundError:
                if not ignore_missing_paths:
                    raise
        elif type_ == "ini":
            try:
                instances.append(config_from_ini(*config_[1:], **default_kwargs))
            except FileNotFoundError:
                if not ignore_missing_paths:
                    raise
        elif type_ == "dotenv":
            try:
                instances.append(config_from_dotenv(*config_[1:], **default_kwargs))
            except FileNotFoundError:
                if not ignore_missing_paths:
                    raise
        elif type_ == "path":
            try:
                instances.append(config_from_path(*config_[1:], **default_kwargs))
            except FileNotFoundError:
                if not ignore_missing_paths:
                    raise
        else:
            raise ValueError(f'Unknown configuration type "{type_}"')

    return ConfigurationSet(
        *instances, interpolate=interpolate, interpolate_type=interpolate_type
    )


class EnvConfiguration(Configuration):
    """Configuration from Environment variables."""

    def __init__(
        self,
        prefix: str,
        separator: str = "__",
        *,
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):
        """
        Constructor.

        :param prefix: prefix to filter environment variables with
        :param separator: separator to replace by dots
        :param lowercase_keys: whether to convert every key to lower case.
        """
        self._prefix = prefix
        self._separator = separator
        super().__init__(
            {},
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )
        self.reload()

    def reload(self) -> None:
        """Reload the environment values."""
        result = {}
        for key, value in os.environ.items():
            if not key.startswith(self._prefix + self._separator):
                continue
            result[
                key[len(self._prefix) :].replace(self._separator, ".").strip(".")
            ] = value
        super().__init__(
            result,
            lowercase_keys=self._lowercase,
            interpolate=self._interpolate,
            interpolate_type=self._interpolate_type,
        )


def config_from_env(
    prefix: str,
    separator: str = "__",
    *,
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`EnvConfiguration` instance from environment variables.

    :param prefix: prefix to filter environment variables with
    :param separator: separator to replace by dots
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return EnvConfiguration(
        prefix,
        separator,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


class PathConfiguration(Configuration):
    """Configuration from a filesytem path."""

    def __init__(
        self,
        path: str,
        remove_level: int = 1,
        *,
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):
        """
        Constructor.

        :param path: path to read from
        :param remove_level: how many levels to remove from the resulting config
        :param lowercase_keys: whether to convert every key to lower case.
        """
        self._path = path
        self._remove_level = remove_level
        super().__init__(
            {},
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )
        self.reload()

    def reload(self) -> None:
        """Reload the path."""
        path = os.path.normpath(self._path)
        if not os.path.exists(path) or not os.path.isdir(path):
            raise FileNotFoundError()

        dotted_path_levels = len(path.split("/"))
        files_keys = (
            (
                os.path.join(x[0], y),
                ".".join(
                    (x[0].split("/") + [y])[(dotted_path_levels + self._remove_level) :]
                ),
            )
            for x in os.walk(path)
            for y in x[2]
            if not x[0].split("/")[-1].startswith("..")
        )

        result = {}
        for filename, key in files_keys:
            result[key] = open(filename).read()

        super().__init__(
            result,
            lowercase_keys=self._lowercase,
            interpolate=self._interpolate,
            interpolate_type=self._interpolate_type,
        )


def config_from_path(
    path: str,
    remove_level: int = 1,
    *,
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`Configuration` instance from filesystem path.

    :param path: path to read from
    :param remove_level: how many levels to remove from the resulting config
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return PathConfiguration(
        path,
        remove_level,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


class FileConfiguration(Configuration):
    """Configuration from a file input."""

    def __init__(
        self,
        data: Union[str, TextIO],
        read_from_file: bool = False,
        *,
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):
        """
        Constructor.

        :param data: path to a config file, or its contents
        :param read_from_file: whether to read from a file path or to interpret
               the :attr:`data` as the contents of the file.
        :param lowercase_keys: whether to convert every key to lower case.
        """
        super().__init__(
            {},
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )
        self._reload(data, read_from_file)
        self._data = data if read_from_file and isinstance(data, str) else None

    def _reload(
        self, data: Union[str, TextIO], read_from_file: bool = False
    ) -> None:  # pragma: no cover
        raise NotImplementedError()

    def reload(self) -> None:
        """Reload the configuration."""
        if self._data:  # pragma: no branch
            self._reload(self._data, True)


class JSONConfiguration(FileConfiguration):
    """Configuration from a JSON input."""

    def _reload(self, data: Union[str, TextIO], read_from_file: bool = False) -> None:
        """Reload the JSON data."""
        if read_from_file:
            if isinstance(data, str):
                result = json.load(open(data, "rt"))
            else:
                result = json.load(data)
        else:
            result = json.loads(cast(str, data))
        self._config = self._flatten_dict(result)


def config_from_json(
    data: Union[str, TextIO],
    read_from_file: bool = False,
    *,
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`Configuration` instance from a JSON file.

    :param data: path to a JSON file or contents
    :param read_from_file: whether to read from a file path or to interpret
           the :attr:`data` as the contents of the JSON file.
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return JSONConfiguration(
        data,
        read_from_file,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


class INIConfiguration(FileConfiguration):
    """Configuration from an INI file input."""

    def __init__(
        self,
        data: Union[str, TextIO],
        read_from_file: bool = False,
        *,
        section_prefix: str = "",
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):
        self._section_prefix = section_prefix
        super().__init__(
            data=data,
            read_from_file=read_from_file,
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )

    def _reload(self, data: Union[str, TextIO], read_from_file: bool = False) -> None:
        """Reload the INI data."""
        import configparser

        lowercase = self._lowercase

        class ConfigParser(configparser.RawConfigParser):
            def optionxform(self, optionstr: str) -> str:
                return super().optionxform(optionstr) if lowercase else optionstr

        if read_from_file:
            if isinstance(data, str):
                data = open(data, "rt").read()
            else:
                data = data.read()
        data = cast(str, data)
        cfg = ConfigParser()
        cfg.read_string(data)
        result = {
            section[len(self._section_prefix) :] + "." + k: v
            for section, values in cfg.items()
            for k, v in values.items()
            if section.startswith(self._section_prefix)
        }
        self._config = self._flatten_dict(result)


def config_from_ini(
    data: Union[str, TextIO],
    read_from_file: bool = False,
    *,
    section_prefix: str = "",
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`Configuration` instance from an INI file.

    :param data: path to an INI file or contents
    :param read_from_file: whether to read from a file path or to interpret
           the :attr:`data` as the contents of the INI file.
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return INIConfiguration(
        data,
        read_from_file,
        section_prefix=section_prefix,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


class DotEnvConfiguration(FileConfiguration):
    """Configuration from a .env type file input."""

    def _reload(self, data: Union[str, TextIO], read_from_file: bool = False) -> None:
        """Reload the .env data."""
        if read_from_file:
            if isinstance(data, str):
                data = open(data, "rt").read()
            else:
                data = data.read()
        data = cast(str, data)
        result: Dict[str, Any] = dict(
            (y.strip() for y in x.split("=", 1))  # type: ignore
            for x in data.splitlines()
            if x
        )
        self._config = self._flatten_dict(result)


def config_from_dotenv(
    data: Union[str, TextIO],
    read_from_file: bool = False,
    *,
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`Configuration` instance from a .env type file.

    :param data: path to a .env type file or contents
    :param read_from_file: whether to read from a file path or to interpret
           the :attr:`data` as the contents of the INI file.
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return DotEnvConfiguration(
        data,
        read_from_file,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


class PythonConfiguration(Configuration):
    """Configuration from a python module."""

    def __init__(
        self,
        module: Union[str, ModuleType],
        prefix: str = "",
        separator: str = "_",
        *,
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):
        """
        Constructor.

        :param module: a module or path string
        :param prefix: prefix to use to filter object names
        :param separator: separator to replace by dots
        :param lowercase_keys: whether to convert every key to lower case.
        """
        if isinstance(module, str):
            if module.endswith(".py"):
                import importlib.util
                from importlib import machinery

                spec = cast(
                    machinery.ModuleSpec,
                    importlib.util.spec_from_file_location(module, module),
                )
                module = importlib.util.module_from_spec(spec)
                spec.loader = cast(InspectLoader, spec.loader)
                spec.loader.exec_module(module)
            else:
                import importlib

                module = importlib.import_module(module)
        self._module = module
        self._prefix = prefix
        self._separator = separator
        super().__init__(
            {},
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )
        self.reload()

    def reload(self) -> None:
        """Reload the path."""
        variables = [
            x
            for x in dir(self._module)
            if not x.startswith("__") and x.startswith(self._prefix)
        ]
        result = {
            k[len(self._prefix) :]
            .replace(self._separator, ".")
            .strip("."): getattr(self._module, k)
            for k in variables
        }
        super().__init__(
            result,
            lowercase_keys=self._lowercase,
            interpolate=self._interpolate,
            interpolate_type=self._interpolate_type,
        )


def config_from_python(
    module: Union[str, ModuleType],
    prefix: str = "",
    separator: str = "_",
    *,
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`Configuration` instance from the objects in a Python module.

    :param module: a module or path string
    :param prefix: prefix to use to filter object names
    :param separator: separator to replace by dots
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return PythonConfiguration(
        module,
        prefix,
        separator,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


def config_from_dict(
    data: dict,
    *,
    lowercase_keys: bool = False,
    interpolate: InterpolateType = False,
    interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
) -> Configuration:
    """
    Create a :class:`Configuration` instance from a dictionary.

    :param data: dictionary with string keys
    :param lowercase_keys: whether to convert every key to lower case.
    :param interpolate: whether to apply string interpolation when looking for items
    :return: a :class:`Configuration` instance
    """
    return Configuration(
        data,
        lowercase_keys=lowercase_keys,
        interpolate=interpolate,
        interpolate_type=interpolate_type,
    )


def create_path_from_config(
    path: str, cfg: Configuration, remove_level: int = 1
) -> Configuration:
    """
    Output a path configuration from a :class:`Configuration` instance.

    :param path: path to create the config files in
    :param cfg: :class:`Configuration` instance
    :param remove_level: how many levels to remove
    """
    import os.path

    assert os.path.isdir(path)

    d = cfg.as_dict()
    for k, v in d.items():
        with open(os.path.join(path, k), "wb") as f:
            f.write(str(v).encode())

        cfg = config_from_path(path, remove_level=remove_level)
    return cfg


if yaml is not None:  # pragma: no branch

    class YAMLConfiguration(FileConfiguration):
        """Configuration from a YAML input."""

        def _reload(
            self, data: Union[str, TextIO], read_from_file: bool = False
        ) -> None:
            """Reload the YAML data."""
            if read_from_file and isinstance(data, str):
                loaded = yaml.load(open(data, "rt"), Loader=yaml.FullLoader)
            else:
                loaded = yaml.load(data, Loader=yaml.FullLoader)
            if not isinstance(loaded, dict):
                raise ValueError("Data should be a dictionary")
            self._config = self._flatten_dict(loaded)

    def config_from_yaml(
        data: Union[str, TextIO],
        read_from_file: bool = False,
        *,
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ) -> Configuration:
        """
        Return a Configuration instance from YAML files.

        :param data: string or file
        :param read_from_file: whether `data` is a file or a YAML formatted string
        :param lowercase_keys: whether to convert every key to lower case.
        :param interpolate: whether to apply string interpolation when looking for items
        :return: a Configuration instance
        """
        return YAMLConfiguration(
            data,
            read_from_file,
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )


if toml is not None:  # pragma: no branch

    class TOMLConfiguration(FileConfiguration):
        """Configuration from a TOML input."""

        def __init__(
            self,
            data: Union[str, TextIO],
            read_from_file: bool = False,
            *,
            section_prefix: str = "",
            lowercase_keys: bool = False,
            interpolate: InterpolateType = False,
            interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
        ):
            self._section_prefix = section_prefix
            super().__init__(
                data=data,
                read_from_file=read_from_file,
                lowercase_keys=lowercase_keys,
                interpolate=interpolate,
                interpolate_type=interpolate_type,
            )

        def _reload(
            self, data: Union[str, TextIO], read_from_file: bool = False
        ) -> None:
            """Reload the TOML data."""
            if read_from_file:
                if isinstance(data, str):
                    loaded = toml.load(open(data, "rt"))
                else:
                    loaded = toml.load(data)
            else:
                data = cast(str, data)
                loaded = toml.loads(data)
            loaded = cast(dict, loaded)

            result = {
                k[len(self._section_prefix) :]: v
                for k, v in self._flatten_dict(loaded).items()
                if k.startswith(self._section_prefix)
            }

            self._config = result

    def config_from_toml(
        data: Union[str, TextIO],
        read_from_file: bool = False,
        *,
        section_prefix: str = "",
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ) -> Configuration:
        """
        Return a Configuration instance from TOML files.

        :param data: string or file
        :param read_from_file: whether `data` is a file or a TOML formatted string
        :param lowercase_keys: whether to convert every key to lower case.
        :param interpolate: whether to apply string interpolation when looking for items
        :return: a Configuration instance
        """
        return TOMLConfiguration(
            data,
            read_from_file,
            section_prefix=section_prefix,
            lowercase_keys=lowercase_keys,
            interpolate=interpolate,
            interpolate_type=interpolate_type,
        )



from typing import Optional

from ...._object_definition import ObjectDefinition
from ._enums import StandardTurnPeriod


class TurnAdjustment(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    standard_turn_period : StandardTurnPeriod, optional

    """

    def __init__(
        self,
        standard_turn_period: Optional[StandardTurnPeriod] = None,
    ) -> None:
        super().__init__()
        self.standard_turn_period = standard_turn_period

    @property
    def standard_turn_period(self):
        """
        :return: enum StandardTurnPeriod
        """
        return self._get_enum_parameter(StandardTurnPeriod, "standardTurnPeriod")

    @standard_turn_period.setter
    def standard_turn_period(self, value):
        self._set_enum_parameter(StandardTurnPeriod, "standardTurnPeriod", value)



import threading
from types import SimpleNamespace
from typing import Optional
import sys

import pytest
from mock.mock import patch

from refinitiv.data import OpenState
from refinitiv.data.delivery._stream import (
    OMMStream,
    StreamConnection,
    get_cxn_cfg_provider,
)
from refinitiv.data.delivery._stream._stream_cxn_cache import (
    stream_cxn_cache,
)
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.delivery.stream import raw_infos
from tests.unit.delivery.stream.conftest import StubWebSocketApp

# --------------------------------------------------------------------------------------
#
# TEST CASES
#
# --------------------------------------------------------------------------------------

"""
Given opened Stream
When network failure happens
Then Cxn is reconnecting, the Stream is opened
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_stream_does_not_close_when_cxn_get_network_failure_and_starts_reconnect():
    cxn: Optional["StreamConnection"] = None
    ws = None
    stream_cxn_cache.cxn_created.clear()

    def get_cxn_ready():
        nonlocal cxn, ws
        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        assert cxn.is_connecting is True, cxn.state

        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    session = StubSession(
        is_open=True,
        response=StubResponse(raw_infos.data),
        stream_auto_reconnection=True,
    )

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=func_name, target=get_cxn_ready, daemon=True).start()
    omm_stream.open()

    assert omm_stream.open_state is OpenState.Opened

    ws.cmd("on_close")
    cxn._listener_created.wait()

    assert cxn.is_connecting is True, cxn.state

    assert omm_stream.open_state is OpenState.Opened

    assert stream_cxn_cache.get_cxns(session) != []


"""
Given opened Stream
When network failure happens
Then Cxn starts cycling infos
And after of all waits some seconds
"""


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_cxn_cycle_infos_and_waits_after_iteration_of_all():
    cxn, ws = None, None
    cxn_ready = threading.Event()
    testing_num_reconnect = 0
    timer_starts = threading.Event()
    stream_cxn_cache.cxn_created.clear()

    def get_cxn_ready():
        nonlocal cxn, ws
        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws

        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    session = StubSession(
        is_open=True,
        response=StubResponse(raw_infos.data),
        stream_auto_reconnection=True,
    )
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    timer_mock = SimpleNamespace()
    timer_mock.wait = lambda *args, **kwargs: timer_starts.set()
    cxn._timer = timer_mock
    expected_num_reconnect = cxn._config.num_infos

    while not timer_starts.is_set():
        ws.cmd("on_close")  # network failure
        cxn._listener_created.wait()  # wait start connecting
        ws = cxn._listener.ws
        testing_num_reconnect += 1

    assert cxn.is_connecting is True, cxn.state

    assert expected_num_reconnect == testing_num_reconnect


"""
Given opened Stream
And network failure happens
When Cxn cycling all infos
And waits from sequence
Then Timeout is the last time in the sequence
"""


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_when_cxn_cycle_infos_timer_get_timeout_in_sequence():
    cxn, ws = None, None
    cxn_ready = threading.Event()
    testing_delay = None
    testing_num_timer_starts = 0
    expected_num_timer_starts = 5
    timer_starts = threading.Event()
    stream_cxn_cache.cxn_created.clear()

    def get_cxn_ready():
        nonlocal cxn, ws
        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws

        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    session = StubSession(
        is_open=True,
        response=StubResponse(raw_infos.data),
        stream_auto_reconnection=True,
    )
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    def timer_wait_mock(timeout):
        nonlocal testing_num_timer_starts, testing_delay
        testing_delay = timeout
        testing_num_timer_starts += 1
        timer_starts.set()

    timer_mock = SimpleNamespace()
    timer_mock.wait = timer_wait_mock
    cxn._timer = timer_mock

    delays = [5, 10, 15, 60, 60]
    for idx in range(expected_num_timer_starts):
        while not timer_starts.is_set():
            ws.cmd("on_close")  # network failure
            cxn._listener_created.wait()  # wait start connecting
            ws = cxn._listener.ws
        timer_starts.clear()
        expected_delay = delays[idx]
        assert expected_delay == testing_delay

    assert cxn.is_connecting is True, cxn.state

    assert testing_num_timer_starts == expected_num_timer_starts


"""
Given opened Stream
And network failure happens
When Cxn cycling all infos
And waits from sequence
And network restored
Then delays reset
And timeout take the first of the sequence
"""


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_when_cxn_cycle_infos_and_restore_network():
    ws, cxn = None, None
    testing_delay = None
    cxn_ready = threading.Event()
    timer_starts = threading.Event()
    stream_cxn_cache.cxn_created.clear()

    def get_cxn_ready():
        nonlocal cxn, ws
        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws

        ws.cmd("on_open")
        message = '[{"State": {"Stream": "Open"}}]'
        ws.cmd("on_message", message)
        cxn_ready.set()

    def send():
        cxn_ready.wait()
        message = '[{"Type": "Refresh", "Complete": true, "ID": 5}]'
        ws.cmd("on_message", message)

    session = StubSession(
        is_open=True,
        response=StubResponse(raw_infos.data),
        stream_auto_reconnection=True,
    )
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=get_cxn_ready, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=send, daemon=True).start()
    omm_stream.open()

    def timer_wait_mock(timeout):
        nonlocal testing_delay
        testing_delay = timeout
        timer_starts.set()

    timer_mock = SimpleNamespace()
    timer_mock.wait = timer_wait_mock
    cxn._timer = timer_mock

    delays = [5, 10, 15, 60, 60]
    for idx in range(5):
        while not timer_starts.is_set():
            ws.cmd("on_close")  # network failure
            cxn._listener_created.wait()  # wait start connecting
            ws = cxn._listener.ws
        timer_starts.clear()
        expected_delay = delays[idx]
        assert expected_delay == testing_delay

    reset_called = False

    def delays_reset_mock():
        nonlocal reset_called
        reset_called = True
        origin_delays_reset()

    origin_delays_reset = cxn._delays.reset
    cxn._delays.reset = delays_reset_mock

    ws.cmd("on_open")  # network restore

    assert reset_called is True

    assert cxn.is_connecting is True, cxn.state


"""
Given the Stream is connecting
When network failure happens
Then Cxn is reconnecting, the Stream is pending
"""


# Stream perspective


@pytest.mark.skipif(reason="hangs")
@patch("websocket.WebSocketApp", new=StubWebSocketApp)
def test_stream_does_not_close_when_cxn_get_network_failure_and_starts_reconnect():
    cxn, ws = None, None
    open_event = threading.Event()
    stream_cxn_cache.cxn_created.clear()

    def ws_open():
        nonlocal cxn, ws

        stream_cxn_cache.cxn_created.wait()
        cxn, *_ = stream_cxn_cache.get_cxns(session)
        cxn._listener_created.wait()
        ws = cxn._listener.ws
        ws.cmd("on_open")
        open_event.set()

    def ws_close():
        nonlocal ws
        open_event.wait()
        ws.cmd("on_close")  # network failure
        cxn._listener_created.wait()  # wait start connecting
        ws = cxn._listener.ws

        assert cxn.is_connecting is True, cxn.state
        assert omm_stream.open_state is OpenState.Pending

        # teardown
        cxn._is_auto_reconnect = False
        ws.cmd("on_close")
        omm_stream._stream._opened.set()

    session = StubSession(
        is_open=True,
        response=StubResponse(raw_infos.data),
        stream_auto_reconnection=True,
    )
    cxn_cfg_provider = get_cxn_cfg_provider(session._get_session_cxn_type())
    cxn_cfg_provider.start_connecting()

    omm_stream = OMMStream(session=session, name="name")
    func_name = sys._getframe().f_code.co_name
    threading.Thread(name=f"{func_name}-1", target=ws_open, daemon=True).start()
    threading.Thread(name=f"{func_name}-2", target=ws_close, daemon=True).start()
    omm_stream.open()


# --------------------------------------------------------------------------------------



from refinitiv.data.content.ipa.surfaces import eti


def surface_eti_definition():
    definition = eti.Definition(
        underlying_definition={"instrumentCode": "BNPP.PA@RIC"},
        surface_tag="1",
        surface_layout=eti.SurfaceLayout(
            format=eti.Format.MATRIX, y_point_count=10
        ),
        surface_parameters=eti.EtiCalculationParams(
            price_side=eti.PriceSide.MID,
            volatility_model=eti.VolatilityModel.SVI,
            x_axis=eti.Axis.DATE,
            y_axis=eti.Axis.STRIKE,
        ),
    )
    return definition


def invalid_surface_eti_definition():
    definition = eti.Definition(
        underlying_definition={"instrumentCode": "BNPP.PA@RIC1"},
        surface_tag="1",
        surface_layout=eti.SurfaceLayout(
            format=eti.Format.MATRIX, y_point_count=10
        ),
        surface_parameters=eti.EtiCalculationParams(
            price_side=eti.PriceSide.MID,
            volatility_model=eti.VolatilityModel.SVI,
            x_axis=eti.Axis.DATE,
            y_axis=eti.Axis.STRIKE,
        ),
    )
    return definition



from ..content.historical_pricing._enums import EventTypes


EVENTS_INTERVALS = {
    "tick": {"event_types": None, "adc": "D"},
    "tas": {"event_types": EventTypes.TRADE, "adc": "D"},
    "taq": {"event_types": EventTypes.QUOTE, "adc": "D"},
}

INTERVALS = {
    **EVENTS_INTERVALS,
    "minute": {"pricing": "PT1M", "adc": "D"},
    "1min": {"pricing": "PT1M", "adc": "D"},
    "5min": {"pricing": "PT5M", "adc": "D"},
    "10min": {"pricing": "PT10M", "adc": "D"},
    "30min": {"pricing": "PT30M", "adc": "D"},
    "60min": {"pricing": "PT60M", "adc": "D"},
    "hourly": {"pricing": "PT1H", "adc": "D"},
    "1h": {"pricing": "PT1H", "adc": "D"},
    "daily": {"pricing": "P1D", "adc": "D"},
    "1d": {"pricing": "P1D", "adc": "D"},
    "1D": {"pricing": "P1D", "adc": "D"},
    "7D": {"pricing": "P7D", "adc": "W"},
    "7d": {"pricing": "P7D", "adc": "W"},
    "weekly": {"pricing": "P1W", "adc": "W"},
    "1W": {"pricing": "P1W", "adc": "W"},
    "monthly": {"pricing": "P1M", "adc": "M"},
    "1M": {"pricing": "P1M", "adc": "M"},
    "quarterly": {"pricing": "P3M", "adc": "CQ"},
    "3M": {"pricing": "P3M", "adc": "CQ"},
    "6M": {"pricing": "P6M", "adc": "CS"},
    "yearly": {"pricing": "P1Y", "adc": "CY"},
    "12M": {"pricing": "P1Y", "adc": "CY"},
    "1Y": {"pricing": "P1Y", "adc": "CY"},
}

NON_INTRA_DAY_INTERVALS = {
    "daily",
    "1d",
    "1D",
    "weekly",
    "7D",
    "7d",
    "1W",
    "monthly",
    "1M",
    "quarterly",
    "3M",
    "yearly",
    "1Y",
}



import os

import pytest

import refinitiv.data as rd
from refinitiv.data.delivery import cfs
from tests.integration.conftest import create_platform_session


def check_file_in_folder(file_name, path_to_folder):
    files_list = os.listdir(path_to_folder)
    assert len(files_list) == 1
    assert file_name in files_list
    size = os.stat(os.path.join(path_to_folder, files_list[0])).st_size
    assert size > 0, f"File size == 0 for file {file_name}"


def check_file_in_working_folder(file_name):
    files_list = os.listdir(r"./")
    assert file_name in files_list
    file = list(filter(lambda file: file == file_name, files_list))[0]
    assert os.stat(file).st_size > 0, f"File size == 0 for file {file_name}"


@pytest.fixture(scope="module", autouse=True)
def get_smallest_file_from_bucket():
    session = create_platform_session()
    session.open()
    rd.session.set_default(session)

    response = cfs.file_sets.Definition(bucket="bulk-ESG", page_size=5).get_data()
    result = []
    for fileset in response.data.file_sets:
        for file in fileset:
            result.append(file)
    sorted_files = list(sorted(result, key=lambda file: file.file_size_in_bytes))
    first_file_set_id = sorted_files[0].fileset_id

    first_smallest_files = cfs.files.Definition(first_file_set_id).get_data()
    session.close()

    return first_smallest_files.data.files[0]



from unittest.mock import MagicMock, call

import pytest

import refinitiv.data as rd
from refinitiv.data import OpenState
from refinitiv.data._fin_coder_layer.session import (
    _open_session,
    set_default,
)


@pytest.fixture
def tools():
    def_obj_mock = MagicMock()
    def_obj_mock.get_session.return_value = MagicMock(spec=rd.session.Session)

    tools = {
        "_definition": MagicMock(return_value=def_obj_mock),
        "_load_config": MagicMock(),
        "_config_object": MagicMock(),
        "_set_default": set_default,
    }

    return tools


def test_private_open_with_defaults(tools):
    _name = "default"
    tools["_config_object"].get_param.return_value = _name

    session = _open_session(config_path=None, app_key=None, config_name=None, **tools)
    default_session = rd.session.get_default()

    tools["_definition"].assert_called_once_with(name=_name)
    assert session is default_session


def test_private_open_with_name(tools):
    _name = "MY"

    session = _open_session(config_path=_name, app_key=None, config_name=None, **tools)
    default_session = rd.session.get_default()

    tools["_definition"].assert_called_once_with(name=_name)
    assert session is default_session


def test_private_open_with_app_key(tools):
    app_key = "fooo"
    _name = "MY"

    tools["_config_object"].get_param.return_value = _name

    session = _open_session(
        config_path=None, app_key=app_key, config_name=None, **tools
    )
    default_session = rd.session.get_default()

    tools["_config_object"].get_param.assert_has_calls(
        [call("sessions.default"), call(f"sessions.{_name}")]
    )
    tools["_config_object"].set_param.assert_called_once_with(
        param=f"sessions.{_name}.app-key", value=app_key
    )
    assert session is default_session


def test_private_open_with_name_and_app_key(tools):
    app_key = "fooo2"
    _name = "MY.s"

    session = _open_session(
        config_path=_name, app_key=app_key, config_name=None, **tools
    )
    default_session = rd.session.get_default()

    tools["_config_object"].set_param.assert_called_once_with(
        param=f"sessions.{_name}.app-key", value=app_key
    )
    assert session is default_session


def test_private_open_with_config(tools):
    config_name = "foo"

    session = _open_session(
        config_path=None, app_key=None, config_name=config_name, **tools
    )
    default_session = rd.session.get_default()

    tools["_load_config"].assert_called_once_with(config_name)
    assert session is default_session


def test_close_session(tools):
    session = _open_session(config_path=None, app_key=None, config_name=None, **tools)

    session.info = MagicMock()
    session.open_state = OpenState.Opened
    rd.close_session()

    session.close.assert_called_once()


def test_bad_session_name(tools):
    tools["_config_object"] = rd.get_config()

    with pytest.raises(NameError):
        _open_session(
            config_path="desktop.invalid_session_name",
            app_key="somekey",
            config_name=None,
            **tools,
        )



import os
import sqlite3
from unittest.mock import mock_open
from unittest.mock import patch
from urllib.parse import unquote

import pytest

from refinitiv.data.content.esg.bulk._db_manager import (
    get_query,
    uri_to_path,
    prepare_insert_query,
    prepare_search_query,
)
from refinitiv.data.content.esg.bulk._errors import KeyNotFoundException


def test_db_manager_update_db_does_not_close_connection(db_manager):
    # when
    db_manager.update_db([("", [], {})])

    # then
    try:
        db_manager.commit()
    except sqlite3.ProgrammingError:
        pytest.fail("manager.update_db() closed db connection")


@pytest.mark.parametrize(
    ("test_value", "expected_result"),
    [
        ("filename", "filename"),
        ("file://", ""),
    ],
)
def test_get_query(test_value, expected_result):
    # given
    file_mock = patch("builtins.open", new=mock_open())

    # when
    file_mock.start()
    result = get_query(test_value)
    file_mock.stop()

    # then
    assert result == expected_result


@pytest.mark.parametrize(
    ("input_value", "expected_result"),
    [
        ("file://file.txt", "file.txt"),
        ("file:///file.txt", "/file.txt"),
        ("file:///dir1/file.txt", "/dir1/file.txt"),
        ("file://dir1/file.txt", "dir1/file.txt"),
        ("file://dir1/dir2/file.txt", "dir1/dir2/file.txt"),
        ("file://C:dir1/dir2//file.txt", "C:dir1/dir2/file.txt"),
        ("file://C:dir1//dir2//file.txt", "C:dir1/dir2/file.txt"),
    ],
)
def test_uri_to_path(input_value, expected_result):
    # given

    # when
    test_value = uri_to_path(input_value)

    test_value = os.path.normpath(unquote(test_value))
    expected_result = os.path.normpath(unquote(expected_result))

    # then
    assert test_value == expected_result, test_value


def test_db_manager_update_db_no_filesdata(db_manager):
    # given
    db_manager._actions.get_downloaded = lambda: [1, 2]

    # when
    result = db_manager.update_db("")

    # then
    assert result is None


def test_db_manager_update_db_no_init_file(db_manager):
    # given
    db_manager._actions.get_downloaded = lambda: []

    # when
    # then
    with pytest.raises(FileNotFoundError):
        db_manager.update_db("")


class TestPrepareInsertQuery:
    def test_basic(self):
        assert (
            prepare_insert_query(
                "text #File.attr #{Fields.attr}",
                {"attr": "a"},
                {"attr": "b"},
            )
            == "text b a"
        )

    def test_null(self):
        assert prepare_insert_query("a = #File.attr", {}) == "a = null"

    def test_multilevel_dicts(self):
        # issue with multilevel and nulls:
        # easy to make a typo in multilevel without any errors
        assert (
            prepare_insert_query(
                "text #File.a.b #Fields.c.d",
                {"c": {"d": "x"}},
                {"a": {"b": "y"}},
            )
            == "text y x"
        )

    def test_lists_inside(self):
        assert prepare_insert_query("#Fields.a.0.b", {"a": [{"b": "x"}]}) == "x"
        assert prepare_insert_query("#Fields.a.0", {"a": ["y"]}) == "y"

    def test_nonfile_nonfields(self):
        with pytest.raises(KeyNotFoundException):
            prepare_insert_query("#Filo.attr", {})

    def test_quote_escaping(self):
        # For some reason this applies to fields with 'Name' in it only
        assert (
            prepare_insert_query(
                "#Fields.Name #Fields.LastName #Fields.Address",
                {"Name": "A'b", "LastName": "C'd", "Address": "E'f"},
            )
            == "A''b C''d E'f"
        )


class TestPrepareSearchQuery:
    def test_basic(self):
        # replaces '*' and '#universe#'
        assert (
            prepare_search_query(
                "smth * from #smth# in #{universe}", ["u1", "u2"], ["col1", "col2"]
            )
            == "smth col1, col2 from #smth# in ('u1', 'u2')"
        )

    def test_one_instrument_in_universe(self):
        assert prepare_search_query("in #{universe}", ["u1"]) == "in ('u1')"



__all__ = ("Definition", "RequestMethod", "Response", "Request")

from ._data._endpoint_data import RequestMethod
from ._data._response import Response
from ._data._request import Request
from ._data._endpoint_definition import Definition



import allure
import pytest

from refinitiv.data.discovery import search, search_templates, Views
from refinitiv.data.errors import RDError
from tests.integration.helpers import (
    check_response_value,
    check_if_dataframe_is_not_none,
)


@allure.suite("FinCoder layer")
@allure.feature("FinCoder layer")
@allure.severity(allure.severity_level.NORMAL)
class TestSearch:
    @allure.title("Get search with query using a string")
    @pytest.mark.caseid("C43895895")
    @pytest.mark.parametrize(
        "query, expected_business_entity",
        [("Microsoft", ["QUOTExEQUITY", "ORGANISATION"])],
    )
    def test_search_with_query_using_string_and_get_data(
        self, open_session, query, expected_business_entity
    ):
        response = search(query)

        check_response_value(response, "BusinessEntity", expected_business_entity)

    @allure.title("Get structured search with boost")
    @pytest.mark.caseid("C43895897")
    @pytest.mark.parametrize(
        "view,filter,boost,select,expected_business_entity",
        [
            (
                Views.PHYSICAL_ASSETS,
                "RCSAssetTypeLeaf eq 'oil refinery' and RCSRegionLeaf eq 'Venezuela'",
                "PlantStatus ne 'Normal Operation'",
                "DocumentTitle,BusinessEntity",
                ["PHYSICALASSETxPLANT"],
            )
        ],
    )
    def test_search_with_boost_and_get_data(
        self, open_session, view, filter, boost, select, expected_business_entity
    ):
        response = search(view=view, filter=filter, boost=boost, select=select)

        check_response_value(response, "BusinessEntity", expected_business_entity)

    @allure.title("Get search with invalid params")
    @pytest.mark.caseid("C43895898")
    @pytest.mark.parametrize(
        "view, filter",
        [(Views.PHYSICAL_ASSETS, "Invalid Filters")],
    )
    def test_search_with_invalid_params(self, open_session, view, filter):
        with pytest.raises(RDError) as error:
            search(view=view, filter=filter)

        assert (
            str(error.value)
            == "Error code 400 | Invalid filter: PROPERTY node <Filters> cannot follow PROPERTY node <Invalid>"
        )

    @allure.title("Get search with search template")
    @pytest.mark.caseid("C48966252")
    def test_search_with_search_template(self, load_config, open_session):
        response = search_templates["Equity"].search(
            exchange_name="Frankfurt Stock Exchange",
            navigators="RCSAssetCategoryLeaf(buckets:3, sub:ExchangeName(buckets:2))",
        )

        check_response_value(response, "ExchangeName", ["Frankfurt Stock Exchange"])

    @allure.title("Get search with embedded search template")
    @pytest.mark.caseid("C48966253")
    def test_search_with_embedded_search_template(self, open_session):
        response = search_templates.UnderlyingRICToOption.search(ric="VOD.L")

        check_if_dataframe_is_not_none(response)
        assert "VOD010bB3.EX" in response["RIC"].tolist()

    @allure.title("Create request with invalid search template")
    @pytest.mark.caseid("C48966256")
    def test_invalid_search_template(self, open_session):
        response = search_templates.UnderlyingRICToOption.search(ric="Invalid RIC")

        assert response.empty, f"Response is not empty"



from typing import Union, TYPE_CHECKING
from contextlib import AbstractContextManager, AbstractAsyncContextManager
from .stream_state import StreamState
from ..._core.session import get_default
from ..._open_state import OpenState

if TYPE_CHECKING:
    from ...content.ipa.financial_contracts._quantitative_data_stream import (
        QuantitativeDataStream,
    )
    from ...content.trade_data_service._stream import TradeDataStream
    from ...content.pricing.chain._stream import StreamingChain
    from ...content._universe_streams import _UniverseStreams
    from . import OMMStream, RDPStream

    Stream = Union[
        _UniverseStreams,
        StreamingChain,
        TradeDataStream,
        RDPStream,
        OMMStream,
        QuantitativeDataStream,
    ]

stream_state_to_open_state = {
    StreamState.Unopened: OpenState.Closed,
    StreamState.Opened: OpenState.Opened,
    StreamState.Opening: OpenState.Pending,
    StreamState.Closed: OpenState.Closed,
    StreamState.Closing: OpenState.Pending,
}


class StreamOpenMixin(AbstractContextManager, AbstractAsyncContextManager):
    _stream: "Stream" = None
    _always_use_default_session: bool

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    async def __aenter__(self):
        await self.open_async()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.close()

    def _try_update_session(self):
        if self._always_use_default_session:
            self._session = get_default()
            self._stream.session = self._session

    @property
    def open_state(self) -> OpenState:
        return stream_state_to_open_state.get(self._stream.state)

    def open(self) -> OpenState:
        self._try_update_session()
        self._stream.open()
        return self.open_state

    async def open_async(self) -> OpenState:
        self._try_update_session()
        await self._stream.open_async()
        return self.open_state

    def close(self) -> OpenState:
        self._stream.close()
        return self.open_state


class StreamOpenWithUpdatesMixin(StreamOpenMixin):
    def open(self, with_updates: bool = True) -> OpenState:
        self._try_update_session()
        self._stream.open(with_updates=with_updates)
        return self.open_state

    async def open_async(self, with_updates: bool = True) -> OpenState:
        self._try_update_session()
        await self._stream.open_async(with_updates=with_updates)
        return self.open_state



import refinitiv.data.content.ipa.financial_contracts as rdf
from tests.integration.content.ipa.contracts.swap.conftest import swap_definition_01


def swaption_definition_01():
    definition = rdf.swaption.Definition(
        instrument_tag="BermudanEURswaption",
        settlement_type=rdf.swaption.SwaptionSettlementType.CASH,
        tenor="7Y",
        strike_percent=2.75,
        notional_amount=1000000,
        buy_sell=rdf.swaption.BuySell.BUY,
        swaption_type=rdf.swaption.SwaptionType.PAYER,
        exercise_style=rdf.swaption.ExerciseStyle.BERM,
        bermudan_swaption_definition=rdf.swaption.BermudanSwaptionDefinition(
            exercise_schedule_type=rdf.swaption.ExerciseScheduleType.FLOAT_LEG,
            notification_days=0,
        ),
        underlying_definition=rdf.swap.Definition(tenor="3Y", template="NOK_AB6O"),
        pricing_parameters=rdf.swaption.PricingParameters(
            valuation_date="2020-04-24", nb_iterations=80
        ),
    )

    return definition


def swaption_definition_02():
    definition = rdf.swaption.Definition(
        instrument_tag="myGBPswaption",
        settlement_type=rdf.swaption.SwaptionSettlementType.CASH,
        tenor="5Y",
        strike_percent=2,
        notional_amount=1000000,
        buy_sell=rdf.swaption.BuySell.BUY,
        exercise_style=rdf.swaption.ExerciseStyle.EURO,
        swaption_type=rdf.swaption.SwaptionType.PAYER,
        bermudan_swaption_definition=rdf.swaption.BermudanSwaptionDefinition(
            exercise_schedule_type=rdf.swaption.ExerciseScheduleType.FLOAT_LEG,
            notification_days=0,
        ),
        underlying_definition=rdf.swap.Definition(tenor="5Y", template="OIS_SONIA"),
        pricing_parameters=rdf.swaption.PricingParameters(valuation_date="2021-11-16"),
        fields=[
            "MarketValueInDealCcy",
            "MarketValueInReportCcy",
            "DeltaPercent",
            "GammaPercent",
            "ThetaPercent",
            "ErrorCode",
            "ErrorMessage",
        ],
    )

    return definition


swaption_universe = {
    "instrumentType": "Swaption",
    "instrumentDefinition": {
        "instrumentTag": "BermudanEURswaption",
        "tenor": "7Y",
        "notionalAmount": 1000000,
        "bermudanSwaptionDefinition": {
            "exerciseScheduleType": "FloatLeg",
            "notificationDays": 0,
        },
        "buySell": "Buy",
        "exerciseStyle": "BERM",
        "settlementType": "Cash",
        "swaptionType": "Payer",
        "underlyingDefinition": {"tenor": "3Y", "template": "NOK_AB6O"},
        "strikePercent": 2.75,
    },
    "pricingParameters": {"nbIterations": 80, "valuationDate": "2020-04-24"},
}



from typing import Callable, List, Optional, Any


def is_date_true(param: Any) -> bool:
    return param is not None


class ParamItem:
    def __init__(
        self,
        arg_name: str,
        query_param_name: Optional[str] = None,
        function: Optional[Callable] = None,
        is_true: Callable = None,
    ):
        self.arg_name = arg_name
        self.query_param_name = query_param_name or arg_name
        self.function = function
        self.is_true = lambda param: is_true(param) if is_true else bool(param)


class ValueParamItem(ParamItem):
    def __init__(
        self,
        arg_name: str,
        query_param_name: Optional[str] = None,
        function: Optional[Callable] = None,
        is_true: Callable = None,
    ):
        super().__init__(arg_name, query_param_name, function, is_true)
        self.function = lambda value, *args, **kwargs: function(value)


class SerParamItem(ParamItem):
    def __init__(
        self,
        arg_name: str,
        query_param_name: Optional[str] = None,
        function: Optional[Callable] = None,
        is_true: Callable = None,
    ):
        super().__init__(arg_name, query_param_name, function, is_true)
        self.function = (
            lambda v, *args, **kwargs: v if not hasattr(v, "get_dict") else v.get_dict()
        )


def get_params(params_config: List[ParamItem], *args, **kwargs):
    retval = []
    for item in params_config:
        param = kwargs.get(item.arg_name)
        if not item.is_true(param):
            continue

        if item.function:
            _kwargs = {k: v for k, v in kwargs.items() if k != item.arg_name}
            param = item.function(param, *args, **_kwargs)
        retval.append((item.query_param_name, param))
    return retval



import warnings
from dataclasses import dataclass
from typing import Dict, Iterable, TYPE_CHECKING, List

import pandas as pd

from .._content_data import Data
from .._content_data_provider import ContentDataProvider
from .._content_response_factory import ContentResponseFactory
from .._error_parser import ErrorParser
from ..._tools._dataframe import convert_df_columns_to_datetime, convert_dtypes
from ...delivery._data._data_provider import (
    RequestFactory,
)
from ...delivery.endpoint_request import RequestMethod

if TYPE_CHECKING:
    from ...delivery._data._data_provider import ParsedData


# --------------------------------------------------------------------------------------
#   Response
# --------------------------------------------------------------------------------------


def _get_unique_keys(list_of_dict: Iterable[Dict]) -> list:
    keys = []
    unique_keys = set()

    for item in list_of_dict:
        unique_keys.update(item)
        if len(keys) < len(item):
            keys = list(item.keys())

    diff_set = unique_keys - set(keys)
    if diff_set:
        keys += diff_set

    return keys


def discovery_search_build_df(raw: dict, **kwargs) -> pd.DataFrame:
    if "Hits" not in raw:
        return pd.DataFrame()

    hits = raw["Hits"]
    keys = _get_unique_keys(hits)
    hit_dataframe = {key: [item.get(key) for item in hits] for key in keys}

    if hit_dataframe:
        df = pd.DataFrame(hit_dataframe)
        convert_df_columns_to_datetime(df, entry="Date", utc=True, delete_tz=True)
        df = convert_dtypes(df)
        return df

    return pd.DataFrame()


def discovery_lookup_build_df(raw: dict, **kwargs) -> pd.DataFrame:
    if "Matches" not in raw:
        return pd.DataFrame()

    match_df = {}
    matches = raw["Matches"]
    matches_list_of_dict = list(matches.values())
    property_names = _get_unique_keys(matches_list_of_dict)

    for match_name, match_value_dict in matches.items():
        for property_name in property_names:
            match_property_df_dict = match_df.setdefault(property_name, {})

            get_name = match_value_dict.get(property_name)
            match_property_df_dict[match_name] = get_name

    if not match_df:
        return pd.DataFrame()

    df = pd.DataFrame(match_df)
    df = convert_dtypes(df)
    return df


#   response data keyword
_ResponsePropertiesName = "Properties"
_ResponseTypeName = "Type"

#       flags
_ResponseFlagNames = [
    _ResponseTypeName,
    # flags
    "Searchable",
    "Sortable",
    "Navigable",
    "Groupable",
    "Exact",
    "Symbol",
]


def _extend_tuple_with_last_element(input_tuple, num_expected_tuple_elements):
    """do extend the input tuple by duplicate last element value to be the
    expected number of tuple elements"""
    num_input_tuple_elements = len(input_tuple)
    return input_tuple + input_tuple[num_input_tuple_elements - 1 :] * (
        num_expected_tuple_elements - num_input_tuple_elements
    )


def _convert_property_attributes(
    property_attribute_to_property_dict_dict,
    property_name,
    ancestor_property_name_type,
    property_attribute_dict,
):
    """convert each property into a dictionary of property attribute
    to property dictionary"""
    #   depth of this property
    this_property_depth = 1

    #   determine this attribute name tuple
    attribute_name_type = ancestor_property_name_type[:] + (property_name,)

    #   check the attribute type is nested or not?
    property_attribute_type = property_attribute_dict[_ResponseTypeName]
    if property_attribute_type == "Nested":
        #   this property is a nested type, recursive convert this property attribute
        #   extract properties of this nested type
        this_properties_of_nested_type = property_attribute_dict[
            _ResponsePropertiesName
        ]

        #   loop over all nested attributes and convert it
        for (
            nested_property_name,
            nested_property_attribute_dict,
        ) in this_properties_of_nested_type.items():
            #   call convert recusivly for nested type
            _convert_property_attributes(
                property_attribute_to_property_dict_dict,
                nested_property_name,
                attribute_name_type,
                nested_property_attribute_dict,
            )

        #   increase property depth by one
        this_property_depth += 1

    #   convert the attribute flags of this property attributes
    #       loop over all possible attributes and convert it
    for property_attribute_name in _ResponseFlagNames:
        #   add properties of properties to dict of dict
        #       fill with None if the properties doesn't exist
        property_dict = property_attribute_to_property_dict_dict.setdefault(
            property_attribute_name, {}
        )
        property_dict[attribute_name_type] = property_attribute_dict.get(
            property_attribute_name, False
        )

    return this_property_depth


def discovery_metadata_build_df(raw: dict, **kwargs) -> pd.DataFrame:
    """parse the metadata response from dict of dict to be a dict of dict tuple"""
    if _ResponsePropertiesName not in raw:
        return pd.DataFrame()

    property_attribute_to_property_dict_dict = {}
    property_name_to_depth_dict = {}

    properties = raw[_ResponsePropertiesName]

    # loop over view metadata and convert it to dict of dict tuple
    for property_name, property_attribute_dict in properties.items():
        # do convert each property attributes into dict of dict
        this_property_depth = _convert_property_attributes(
            property_attribute_to_property_dict_dict,
            property_name,
            (),
            property_attribute_dict,
        )

        # store each property depth as dict
        property_name_to_depth_dict[property_name] = this_property_depth

    ######################################################
    # convert response to be a dataframe format
    # warning OPTIMIZE_ME :: THIS CAN BE OPTIMIZE

    #   determine the maximum depth on nested property
    max_property_depth = max(property_name_to_depth_dict.values())

    # loop over all the propertyAttributeToPropertyDictDict
    # and convert to a dataframe format
    data = {}
    for (
        property_attribute_name,
        property_dict,
    ) in property_attribute_to_property_dict_dict.items():
        for property_key, property_value in property_dict.items():
            #   construct the property attribute for this property
            dataframe_property_dict = data.setdefault(property_attribute_name, {})

            #   construct the key of dataframe property
            dataframe_property_key = _extend_tuple_with_last_element(
                property_key, max_property_depth
            )
            dataframe_property_dict[dataframe_property_key] = property_value

    if len(data) == 0:
        return pd.DataFrame()

    df = pd.DataFrame(data)
    df = convert_dtypes(df)
    return df


class Bucket:
    def __init__(self, data):
        self._data = data
        self._subnavigator = None

    @property
    def label(self):
        return self._data["Label"]

    @property
    def navigator(self):
        if not self._subnavigator:
            for name, value in self._data.items():
                if isinstance(self._data[name], dict):
                    buckets_objects = [
                        Bucket(bucket) for bucket in self._data[name]["Buckets"]
                    ]
                    self._subnavigator = Navigator(buckets_objects, name)
        return self._subnavigator

    @property
    def count(self):
        return self._data["Count"]

    def __getitem__(self, name):
        try:
            return self._data[name]
        except KeyError:
            raise KeyError(f"'{name}' property does not exist.")


@dataclass
class HitsItem:
    BusinessEntity: str
    DocumentTitle: str
    PermID: str
    PI: str
    RIC: str


@dataclass
class Navigator:
    buckets: List
    name: str


class SearchData(Data):
    _navigators = None
    _hits = None
    _total = None

    @property
    def total(self):
        if not self._total:
            self._total = self.raw.get("Total")
        return self._total

    @property
    def hits(self):
        if not self._hits:
            self._hits = [
                HitsItem(
                    item.get("BusinessEntity"),
                    item.get("DocumentTitle"),
                    item.get("PermID"),
                    item.get("PI"),
                    item.get("RIC"),
                )
                for item in self.raw.get("Hits", [])
            ]
        return self._hits

    @property
    def navigators(self):
        if not self._navigators:
            self._navigators = {
                name: Navigator([Bucket(bucket) for bucket in buckets["Buckets"]], name)
                for name, buckets in self.raw.get("Navigators", {}).items()
            }

        return self._navigators


class SearchResponseFactory(ContentResponseFactory):
    def create_success(self, parsed_data: "ParsedData", **kwargs):
        inst = super().create_success(parsed_data, **kwargs)
        skipped = parsed_data.content_data.get("Skipped", 0)
        total = parsed_data.content_data.get("Total", 0)
        hits = len(parsed_data.content_data.get("Hits", []))
        query = kwargs.get("query", "")
        if total > hits + skipped == 10000:
            warnings.warn(
                f"Search result for {query} not full.\n"
                f"Result is maxed at 10000 while the total is {total} rows.\n"
                f"Requested - {hits}, skipped - {skipped} rows."
            )

        inst.total = total
        return inst


# --------------------------------------------------------------------------------------
#   Request
# --------------------------------------------------------------------------------------


class BaseSearchRequestFactory(RequestFactory):
    defn_params_by_body_params = {}

    def get_body_parameters(self, *_, **kwargs) -> dict:
        return {
            self.defn_params_by_body_params[param_name]: param_value
            for param_name, param_value in kwargs.items()
            if param_name in self.defn_params_by_body_params and param_value is not None
        }

    def get_request_method(self, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class SearchRequestFactory(BaseSearchRequestFactory):
    defn_params_by_body_params = {
        "boost": "Boost",
        "features": "Features",
        "filter": "Filter",
        "group_by": "GroupBy",
        "group_count": "GroupCount",
        "navigators": "Navigators",
        "order_by": "OrderBy",
        "query": "Query",
        "scope": "Scope",
        "select": "Select",
        "skip": "Skip",
        "terms": "Terms",
        "top": "Top",
        "view": "View",
    }


class LookupRequestFactory(BaseSearchRequestFactory):
    defn_params_by_body_params = {
        "boost": "Boost",
        "filter": "Filter",
        "scope": "Scope",
        "select": "Select",
        "terms": "Terms",
        "view": "View",
    }


class MetadataRequestFactory(RequestFactory):
    def get_query_parameters(self, *, view, **_):
        return view

    def add_query_parameters(self, url, query_parameters) -> str:
        return f"{url}/{query_parameters}"


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------


search_data_provider = ContentDataProvider(
    request=SearchRequestFactory(),
    response=SearchResponseFactory(data_class=SearchData),
    parser=ErrorParser(),
)

lookup_data_provider = ContentDataProvider(
    request=LookupRequestFactory(),
    parser=ErrorParser(),
)

metadata_data_provider = ContentDataProvider(
    request=MetadataRequestFactory(),
    parser=ErrorParser(),
)



# coding: utf8


from typing import Optional, List

from . import BermudanSwaptionDefinition
from ._enums import (
    BuySell,
    ExerciseStyle,
    PremiumSettlementType,
    SwaptionSettlementType,
    SwaptionType,
)
from ._models import InputFlow
from .. import swap
from .._instrument_definition import InstrumentDefinition
from ..swap._swap_definition import SwapInstrumentDefinition


class SwaptionInstrumentDefinition(InstrumentDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        A user defined string to identify the instrument. it can be used to link output
        results to the instrument definition.limited to 40 characters.only alphabetic,
        numeric and '- _.#=@' characters are supported. optional. no default value
        applies.
    start_date : str, optional
        The date the swaption starts. optional. by default it is derived from the
        tradedate and the day to spot convention of the contract currency.
    end_date : str, optional
        The maturity or expiry date of the instrument's leg. the value is expressed in
        iso 8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g. 2021-01-01t00:00:00z). either
        tenor or enddate must be provided. the default value is valuationdate shifted
        forward by tenor.
    tenor : str, optional
        The code indicating the period between startdate and enddate of the instrument
        (e.g. '6m', '1y'). mandatory, if enddate is not provided. the default value is
        calculated from enddate.
    notional_amount : float, optional
        The notional amount of the instrument. the default value is '1,000,000'.
    bermudan_swaption_definition : BermudanSwaptionDefinition, optional

    buy_sell : BuySell, optional
        The indicator of the deal side. the possible values are:   buy: buying the
        option,   sell: selling/writing the option.  no default value applies.
    exercise_style : ExerciseStyle, optional
        The option style based on its exercise restrictions. the possible values are:
        amer,   euro,   berm.  note: all exercise styles may not apply to certain option
        no default value applies.
    payments : InputFlow, optional
        An array of payments
    premium_settlement_type : PremiumSettlementType, optional
        The cash settlement type of the option premium   spot,   forward.
    settlement_type : SwaptionSettlementType, optional
        The settlement method for options when exercised. the possible values are:
        physical: delivering the underlying asset, or for a swaption, physically
        entering into the underlying swap.    cash: paying out in cash.  the default
        value is 'physical'.
    swaption_type : SwaptionType, optional
        The indicator if the swaption is a payer or a receiver. the possible values are:
        receiver: a right to receive a fixed rate of the underlying swap,   payer: a
        right to pay a fixed rate of the underlying swap.  no default value applies.
    underlying_definition : SwapDefinition, optional

    spread_vs_atm_in_bp : float, optional
        Spread between strike and atm strike, expressed in basis points (bp).
    strike_percent : float, optional
        The set price at which the owner of the option can buy or sell the underlying
        asset. for a swaption, it is the fixed rate of the underlying swap at which the
        owner of the swaption can enter the swap. the value is expressed in percentages.
        by default, fixedratepercent of the underlying swap is used.
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_amount: Optional[float] = None,
        bermudan_swaption_definition: Optional[BermudanSwaptionDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        payments: Optional[List[InputFlow]] = None,
        premium_settlement_type: Optional[PremiumSettlementType] = None,
        settlement_type: Optional[SwaptionSettlementType] = None,
        swaption_type: Optional[SwaptionType] = None,
        underlying_definition: Optional[swap.Definition] = None,
        spread_vs_atm_in_bp: Optional[float] = None,
        strike_percent: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.start_date = start_date
        self.end_date = end_date
        self.tenor = tenor
        self.notional_amount = notional_amount
        self.bermudan_swaption_definition = bermudan_swaption_definition
        self.buy_sell = buy_sell
        self.exercise_style = exercise_style
        self.payments = payments
        self.premium_settlement_type = premium_settlement_type
        self.settlement_type = settlement_type
        self.swaption_type = swaption_type
        self.underlying_definition = underlying_definition
        self.spread_vs_atm_in_bp = spread_vs_atm_in_bp
        self.strike_percent = strike_percent

    @classmethod
    def get_instrument_type(cls):
        return "Swaption"

    @property
    def bermudan_swaption_definition(self):
        """
        :return: object BermudanSwaptionDefinition
        """
        return self._get_object_parameter(
            BermudanSwaptionDefinition, "bermudanSwaptionDefinition"
        )

    @bermudan_swaption_definition.setter
    def bermudan_swaption_definition(self, value):
        self._set_object_parameter(
            BermudanSwaptionDefinition, "bermudanSwaptionDefinition", value
        )

    @property
    def buy_sell(self):
        """
        The side of the deal.
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def exercise_style(self):
        """
        :return: enum ExerciseStyle
        """
        return self._get_enum_parameter(ExerciseStyle, "exerciseStyle")

    @exercise_style.setter
    def exercise_style(self, value):
        self._set_enum_parameter(ExerciseStyle, "exerciseStyle", value)

    @property
    def payments(self):
        """
        An array of payments
        :return: list InputFlow
        """
        return self._get_list_parameter(InputFlow, "payments")

    @payments.setter
    def payments(self, value):
        self._set_list_parameter(InputFlow, "payments", value)

    @property
    def premium_settlement_type(self):
        """
        The cash settlement type of the option premium   spot,   forward.
        :return: enum PremiumSettlementType
        """
        return self._get_enum_parameter(PremiumSettlementType, "premiumSettlementType")

    @premium_settlement_type.setter
    def premium_settlement_type(self, value):
        self._set_enum_parameter(PremiumSettlementType, "premiumSettlementType", value)

    @property
    def settlement_type(self):
        """
        The settlement type of the option if the option is exercised.
        :return: enum SwaptionSettlementType
        """
        return self._get_enum_parameter(SwaptionSettlementType, "settlementType")

    @settlement_type.setter
    def settlement_type(self, value):
        self._set_enum_parameter(SwaptionSettlementType, "settlementType", value)

    @property
    def swaption_type(self):
        """
        The indicator if the swaption is a payer or a receiver. the possible values are:
        receiver: a right to receive a fixed rate of the underlying swap,   payer: a
        right to pay a fixed rate of the underlying swap.  no default value applies.
        :return: enum SwaptionType
        """
        return self._get_enum_parameter(SwaptionType, "swaptionType")

    @swaption_type.setter
    def swaption_type(self, value):
        self._set_enum_parameter(SwaptionType, "swaptionType", value)

    @property
    def underlying_definition(self):
        """
        :return: object SwapDefinition
        """
        return self._get_object_parameter(
            SwapInstrumentDefinition, "underlyingDefinition"
        )

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(
            SwapInstrumentDefinition, "underlyingDefinition", value
        )

    @property
    def end_date(self):
        """
        The maturity or expiry date of the instrument's leg. the value is expressed in
        iso 8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g. 2021-01-01t00:00:00z). either
        tenor or enddate must be provided. the default value is valuationdate shifted
        forward by tenor.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def instrument_tag(self):
        """
        A user defined string to identify the instrument. it can be used to link output
        results to the instrument definition.limited to 40 characters.only alphabetic,
        numeric and '- _.#=@' characters are supported. optional. no default value
        applies.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        self._set_parameter("instrumentTag", value)

    @property
    def notional_amount(self):
        """
        The notional amount of the instrument. The default value is '1,000,000'.
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def spread_vs_atm_in_bp(self):
        """
        Spread between strike and atm strike, expressed in basis points (bp).
        :return: float
        """
        return self._get_parameter("spreadVsAtmInBp")

    @spread_vs_atm_in_bp.setter
    def spread_vs_atm_in_bp(self, value):
        self._set_parameter("spreadVsAtmInBp", value)

    @property
    def start_date(self):
        """
        The date the swaption starts. optional. by default it is derived from the
        tradedate and the day to spot convention of the contract currency.
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def strike_percent(self):
        """
        The set price at which the owner of the option can buy or sell the underlying
        asset. for a swaption, it is the fixed rate of the underlying swap at which the
        owner of the swaption can enter the swap. the value is expressed in percentages.
        by default, fixedratepercent of the underlying swap is used.
        :return: float
        """
        return self._get_parameter("strikePercent")

    @strike_percent.setter
    def strike_percent(self, value):
        self._set_parameter("strikePercent", value)

    @property
    def tenor(self):
        """
        The code indicating the period between startdate and enddate of the instrument
        (e.g. '6m', '1y'). mandatory, if enddate is not provided. the default value is
        calculated from enddate.
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)



# coding: utf8

from typing import TYPE_CHECKING, List, Callable

import pandas as pd

from .._content_data_provider import ContentDataProvider
from .._content_response_factory import ContentResponseFactory
from .._error_parser import ErrorParser
from ..._tools import PRICING_DATETIME_PATTERN
from ..._tools._common import fields_arg_parser, universe_arg_parser, cached_property
from ..._tools._dataframe import convert_df_columns_to_datetime_re, convert_dtypes
from ...delivery._data._data_provider import (
    ContentValidator,
    RequestFactory,
    ValidatorContainer,
)
from ...delivery._stream.stream_cache import StreamCache

if TYPE_CHECKING:
    from ...delivery._data._data_provider import ParsedData


# ---------------------------------------------------------------------------
#   Response factory
# ---------------------------------------------------------------------------


class PriceCache:
    def __init__(self, cache: dict):
        self._cache = cache

    def keys(self):
        return self._cache.keys()

    def values(self):
        return self._cache.values()

    def items(self):
        return self._cache.items()

    def __iter__(self):
        return PricingCacheIterator(self)

    def __getitem__(self, name):
        if name in self.keys():
            return self._cache[name]
        raise KeyError(f"{name} not in PriceCache")

    def __len__(self):
        return len(self.keys())

    def __str__(self):
        return str(self._cache)


class PricingCacheIterator:
    def __init__(self, price_cache: PriceCache):
        self._price_cache = price_cache
        self._universe = list(price_cache.keys())
        self._index = 0

    def __next__(self):
        if self._index < len(self._universe):
            name = self._universe[self._index]
            result = self._price_cache[name]
            self._index += 1
            return result
        raise StopIteration()


def create_price_cache(data: dict, fields) -> PriceCache:
    cache = {}
    for item in data:
        key = item.get("Key")
        if key:
            name = key.get("Name")
            service = key.get("Service")
            status = item.get("State")
            cache[name] = StreamCache(
                name=name,
                fields=fields,
                service=service,
                status=status,
                record=item,
            )
    return PriceCache(cache)


status_code_to_value = {"NotEntitled": "#N/P", "NotFound": "#N/F"}


def pricing_build_df(
    raw: List[dict], universe: list, fields: list, **kwargs
) -> pd.DataFrame:
    """Pricing dataframe builder.
    Args:
        raw (List[dict]): list of raw data to build dataframe.
        universe (list): list of RICs.
        fields (list): list of fields used to build dataframe.
        **kwargs: additional keyword arguments.
    Returns:
        DataFrame: properly created dataframe.
    """
    if not fields:
        fields = list(
            dict.fromkeys(key for item in raw for key in item.get("Fields", {}).keys())
        )

    data = []
    num_fields = len(fields)
    for idx, item in enumerate(raw):
        inst_name = universe[idx]
        if item["Type"] == "Status":
            value = status_code_to_value.get(item["State"]["Code"])
            values = [value] * num_fields
            data.append([inst_name, *values])
        else:
            row = [inst_name]
            for field in fields:
                value = item["Fields"].get(field)
                value = pd.NA if value is None else value
                row.append(value)
            data.append(row)

    df = pd.DataFrame(data=data, columns=["Instrument", *fields])
    convert_df_columns_to_datetime_re(df, PRICING_DATETIME_PATTERN)
    df = convert_dtypes(df)
    return df


class PricingResponseFactory(ContentResponseFactory):
    def create_data_success(self, parsed_data: "ParsedData", **kwargs):
        data = super().create_data_success(parsed_data, **kwargs)
        data.prices = create_price_cache(
            self.get_raw(parsed_data), kwargs.get("fields")
        )
        return data


# ---------------------------------------------------------------------------
#   Request factory
# ---------------------------------------------------------------------------


class PricingRequestFactory(RequestFactory):
    def get_query_parameters(self, *args, **kwargs) -> list:
        query_parameters = []

        #
        # universe
        #
        universe = kwargs.get("universe")
        if universe:
            universe = universe_arg_parser.get_str(universe, delim=",")
            query_parameters.append(("universe", universe))

        #
        # fields
        #
        fields = kwargs.get("fields")
        if fields:
            fields = fields_arg_parser.get_str(fields, delim=",")
            query_parameters.append(("fields", fields))

        return query_parameters


# ---------------------------------------------------------------------------
#   Content data validator
# ---------------------------------------------------------------------------


class PricingContentValidator(ContentValidator):
    @cached_property
    def validators(self) -> List[Callable[["ParsedData"], bool]]:
        return [self.status_is_not_error]


# ---------------------------------------------------------------------------
#   Data provider
# ---------------------------------------------------------------------------

pricing_data_provider = ContentDataProvider(
    request=PricingRequestFactory(),
    response=PricingResponseFactory(),
    parser=ErrorParser(),
    validator=ValidatorContainer(content_validator=PricingContentValidator()),
)



import pytest

from refinitiv.data.errors import RDError
from tests.unit.conftest import StubSession, StubResponse, parametrize_with_test_case
from tests.unit.content.ipa.conftest import ERROR_MESSAGE_TEST_CASES

args_names = "defn,cnt_data,err_msg,err_code"


@parametrize_with_test_case(
    args_names,
    ERROR_MESSAGE_TEST_CASES,
    "zc_curve_error_message",
    "forward_curve_error_message",
    "zc_curves_error_message",
    "surface_cap_error_message",
    "cap_floor_error_message",
    "cds_error_message",
    "options_error_message",
)
def test_error_message(defn, cnt_data, err_msg, err_code):
    # given
    error_match = f"Error code {err_code} | {err_msg}"

    try:
        status_code = int(err_code)
    except (ValueError, TypeError):
        status_code = None

    response = StubResponse(cnt_data, status_code)
    session = StubSession(is_open=True, response=response)
    definition = defn

    # then
    with pytest.raises(RDError, match=error_match) as e:
        # when
        definition.get_data(session)

    assert e.value.code == err_code
    assert e.value.message == err_msg



from typing import Union, TYPE_CHECKING

from .._content_data import Data
from .._intervals import DayIntervalType, Intervals, get_day_interval_type
from ..._tools import hp_universe_parser, validate_types, try_copy_to_list
from ...delivery._data._data_provider import BaseResponse, DataProviderLayer
from ._historical_pricing_request_factory import get_content_type_by_interval

if TYPE_CHECKING:
    from ._enums import OptAdjustments, OptMarketSession
    from ..._types import (
        OptInt,
        ExtendedParams,
        OptDateTime,
        StrStrings,
        StrStrings,
    )


class Definition(DataProviderLayer[BaseResponse[Data]]):
    """
    Summary line of this class that defines parameters for requesting summaries from historical pricing

    Parameters
    ----------
    universe : str or list of str
        The entity universe
    interval : str or Intervals, optional
        The consolidation interval in ISO8601
    start : str or date or datetime or timedelta, optional
        The start date and timestamp of the query in ISO8601 with UTC only
    end : str or date or datetime or timedelta, optional
        The end date and timestamp of the query in ISO8601 with UTC only
    adjustments : list of Adjustments or Adjustments or str, optional
        The adjustment list or Adjustments type
    sessions : list of MarketSession or MarketSession or str, optional
        The list of market session classification or str
    count : int, optional
        The maximum number of data returned. Values range: 1 - 10000
    fields : list, optional
        The list of fields that are to be returned in the response
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content.historical_pricing import summaries
    >>> definition_summaries = summaries.Definition("EUR")
    >>> response = definition_summaries.get_data()

    """

    def __init__(
        self,
        universe: "StrStrings",
        interval: Union[str, Intervals] = None,
        start: "OptDateTime" = None,
        end: "OptDateTime" = None,
        adjustments: "OptAdjustments" = None,
        sessions: "OptMarketSession" = None,
        count: "OptInt" = None,
        fields: "StrStrings" = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        # By default, if interval is not defined, interday default value is requested
        day_interval_type = get_day_interval_type(interval or DayIntervalType.INTER)
        content_type = get_content_type_by_interval(day_interval_type)
        validate_types(count, [int, type(None)], "count")
        universe = try_copy_to_list(universe)
        universe = hp_universe_parser.get_list(universe)
        adjustments = try_copy_to_list(adjustments)
        sessions = try_copy_to_list(sessions)
        fields = try_copy_to_list(fields)

        super().__init__(
            data_type=content_type,
            universe=universe,
            interval=interval,
            start=start,
            end=end,
            adjustments=adjustments,
            sessions=sessions,
            count=count,
            fields=fields,
            extended_params=extended_params,
        )



# coding: utf8
import abc
import collections
from enum import Enum, unique
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ._response import ABCResponse

Error = collections.namedtuple("Error", ["code", "message"])


@unique
class RequestMethod(str, Enum):
    """
    The RESTful Data service can support multiple methods when
    sending requests to a specified endpoint.
       GET : Request data from the specified endpoint.
       POST : Send data to the specified endpoint to create/update a resource.
       DELETE : Request to delete a resource from a specified endpoint.
       PUT : Send data to the specified endpoint to create/update a resource.
    """

    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"

    def __str__(self) -> str:
        return str(self.value)


class ABCData(abc.ABC):
    pass


class EndpointData(ABCData, object):
    def __init__(self, raw: Any, owner_: "ABCResponse" = None, **kwargs):
        self._raw = raw
        self._owner = owner_
        self._kwargs = kwargs

    @property
    def raw(self):
        return self._raw



from html.parser import HTMLParser


class ParseError(Exception):
    pass


class HeadlineHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.data = {}
        self._write_data = False
        self._key = None

    def handle_starttag(self, tag: str, attrs: tuple):
        # attrs example: (("class", "releasedDate"), ("data-version-created-date", "2021-10-03T11:08:19.096Z"))
        attrs_dict = dict(attrs)
        style_class = attrs_dict.get("class")
        if tag == "span" and style_class == "headline":
            self._write_data = True
            self._key = "headline"
        elif tag == "span" and style_class == "source":
            self.data["creator"] = attrs_dict.get("title")
        elif tag == "span" and style_class == "releasedDate":
            self.data["creation_date"] = attrs_dict.get("data-version-created-date")
        elif tag == "p":
            self._write_data = True
            self._key = "text"

    def handle_data(self, data: str):
        if self._write_data:
            if self._key == "text":
                if self.data.get("text") is None:
                    self.data["text"] = []
                self.data["text"].append(data)
            else:
                self.data[self._key] = data
            self._write_data = False

    def error(self, message):
        raise ParseError(message)



from inspect import signature
import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession
import pytest

from tests.unit.conftest import (
    remove_dunder_methods,
    remove_private_attributes,
    has_property_names_in_class,
    get_property_names,
)


def test_ipa_cap_floor_definition_attributes():
    expected_attributes = ["get_data", "get_data_async", "get_stream"]
    testing_attributes = dir(rdf.cap_floor.Definition)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes


def test_ipa_cap_floor_attributes():
    expected_attributes = [
        "AdjustInterestToPaymentDate",
        "AmortizationFrequency",
        "AmortizationItem",
        "AmortizationType",
        "BarrierDefinitionElement",
        "BarrierType",
        "BusinessDayConvention",
        "BuySell",
        "DateRollingConvention",
        "DayCountBasis",
        "Definition",
        "Frequency",
        "IndexConvexityAdjustmentIntegrationMethod",
        "IndexConvexityAdjustmentMethod",
        "IndexResetType",
        "InputFlow",
        "InterestCalculationConvention",
        "PremiumSettlementType",
        "PriceSide",
        "PricingParameters",
        "StubRule",
    ]
    testing_attributes = dir(rdf.cap_floor)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes, testing_attributes


@pytest.mark.parametrize(
    argnames="input_data",
    ids=[
        "Definition",
        "PricingParameters",
        "AmortizationItem",
        "InputFlow",
        "BarrierDefinitionElement",
    ],
    argvalues=[
        (
            rdf.cap_floor._cap_floor_definition.CapFloorInstrumentDefinition,
            {
                "payment_business_days": "payment_business_days",
                "barrier_definition": rdf.cap_floor.BarrierDefinitionElement(),
                "annualized_rebate": True,
                "cap_digital_payout_percent": 123.123,
                "cms_template": "cms_template",
                "floor_digital_payout_percent": 123.123,
                "instrument_tag": "instrument_tag",
                "start_date": "start_date",
                "end_date": "end_date",
                "tenor": "tenor",
                "notional_ccy": "notional_ccy",
                "notional_amount": 6.6,
                "index_name": "index_name",
                "index_tenor": "index_tenor",
                "interest_payment_frequency": rdf.cap_floor.Frequency.ANNUAL,
                "interest_calculation_method": rdf.cap_floor.DayCountBasis.DCB_30_360,
                "payment_business_day_convention": rdf.cap_floor.BusinessDayConvention.BBSW_MODIFIED_FOLLOWING,
                "payment_roll_convention": rdf.cap_floor.DateRollingConvention.LAST,
                "index_reset_frequency": rdf.cap_floor.Frequency.ANNUAL,
                "index_reset_type": rdf.cap_floor.IndexResetType.IN_ADVANCE,
                "index_fixing_lag": 15,
                "amortization_schedule": [rdf.cap_floor.AmortizationItem()],
                "adjust_interest_to_payment_date": rdf.cap_floor.AdjustInterestToPaymentDate.ADJUSTED,
                "buy_sell": rdf.cap_floor.BuySell.BUY,
                "cap_strike_percent": 19.19,
                "floor_strike_percent": 20.20,
                "index_fixing_ric": "index_fixing_ric",
                "stub_rule": rdf.cap_floor.StubRule.MATURITY,
                "is_backward_looking_index": True,
                "payments": [rdf.cap_floor.InputFlow()],
                "is_rfr": True,
                "interest_calculation_convention": rdf.cap_floor.InterestCalculationConvention.BOND_BASIS,
                "is_term_rate": True,
            },
        ),
        (
            rdf.cap_floor.PricingParameters,
            {
                "index_convexity_adjustment_integration_method": rdf.cap_floor.IndexConvexityAdjustmentIntegrationMethod.RIEMANN_SUM,
                "index_convexity_adjustment_method": rdf.cap_floor.IndexConvexityAdjustmentMethod.BLACK_SCHOLES,
                "market_value_in_deal_ccy": 2.2,
                "report_ccy": "str",
                "skip_first_cap_floorlet": True,
                "valuation_date": "valuation_date",
                "market_data_date": "market_data_date",
                "price_side": rdf.cap_floor.PriceSide.LAST,
            },
        ),
        (
            rdf.cap_floor.AmortizationItem,
            {
                "start_date": "start_date",
                "end_date": "end_date",
                "amortization_frequency": rdf.cap_floor.AmortizationFrequency.ONCE,
                "amortization_type": rdf.cap_floor.AmortizationType.NONE,
                "remaining_notional": 1.1,
                "amount": 2.2,
            },
        ),
        (
            rdf.cap_floor.InputFlow,
            {
                "amount": 1.1,
                "premium_settlement_type": rdf.cap_floor.PremiumSettlementType.FORWARD,
                "currency": "currency",
                "date": "date",
            },
        ),
        (
            rdf.cap_floor.BarrierDefinitionElement,
            {
                "barrier_type": rdf.cap_floor.BarrierType.KNOCK_IN,
                "barrier_down_percent": 1.1,
                "barrier_up_percent": 2.2,
                "rebate_down_percent": 3.3,
                "rebate_up_percent": 4.4,
            },
        ),
    ],
)
def test_parameter(input_data):
    cls, kwargs = input_data
    args_names = list(kwargs.keys())
    inst = cls(**kwargs)

    s = signature(cls.__init__)
    # +1 for (self)
    assert len(s.parameters) == (
        len(args_names) + 1
    ), f"csl={cls}-{str(set(get_property_names(cls)) - set(args_names))}"

    assert has_property_names_in_class(cls, args_names), set(args_names) - set(
        get_property_names(cls)
    )

    for k, v in kwargs.items():
        attr = getattr(inst, k)
        assert attr == v, k


def test_request_body():
    # given
    expected_body = {
        "fields": ["1", "2"],
        "universe": [
            {
                "instrumentDefinition": {
                    "endDate": "2032-02-28",
                    "interestCalculationMethod": "Dcb_Actual_Actual",
                    "interestPaymentFrequency": "Annual",
                    "notionalCcy": "USD",
                },
                "extended_params": "extended_params",
                "instrumentType": "CapFloor",
                "pricingParameters": {"marketDataDate": "market_data_date"},
            }
        ],
    }
    definition = rdf.cap_floor.Definition(
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        interest_calculation_method=rdf.cap_floor.DayCountBasis.DCB_ACTUAL_ACTUAL,
        fields=["1", "2"],
        pricing_parameters=rdf.cap_floor.PricingParameters(
            market_data_date="market_data_date"
        ),
        extended_params={"extended_params": "extended_params"},
    )

    # when
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_body = request.json
    assert testing_body == expected_body, expected_body
